/*
 * This is a Gau2Grid automatically generated C file.
 *
 * More details can found at the following repo:
 *   https://github.com/dgasmith/gau2grid
 */

#include <math.h>
#include <stdio.h>
#if defined __clang__
#include <mm_malloc.h>
#elif defined _MSC_VER
#include <malloc.h>
#else
#include <stdlib.h>
#endif

#include "gau2grid.h"
#include "gau2grid_utility.h"
#include "gau2grid_pragma.h"



void gg_collocation_L0_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 1;
    const unsigned long nspherical = 1;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 32 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);
    double* PRAGMA_RESTRICT phi_x_tmp = (double*)ALIGNED_MALLOC(64, 32 * sizeof(double));
    ASSUME_ALIGNED(phi_x_tmp, 64);
    double* PRAGMA_RESTRICT phi_y_tmp = (double*)ALIGNED_MALLOC(64, 32 * sizeof(double));
    ASSUME_ALIGNED(phi_y_tmp, 64);
    double* PRAGMA_RESTRICT phi_z_tmp = (double*)ALIGNED_MALLOC(64, 32 * sizeof(double));
    ASSUME_ALIGNED(phi_z_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Combine blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            // Gaussian derivs (gradients)
            const double SX = S1[i] * xc[i];
            const double SY = S1[i] * yc[i];
            const double SZ = S1[i] * zc[i];
            phi_out[start + i] = S0[i];

            // Gradient AM=0 Component=0
            phi_x_out[start + i] = SX;
            phi_y_out[start + i] = SY;
            phi_z_out[start + i] = SZ;
        }
    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);
    ALIGNED_FREE(phi_x_tmp);
    ALIGNED_FREE(phi_y_tmp);
    ALIGNED_FREE(phi_z_tmp);

}

void gg_collocation_L1_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 3;
    const unsigned long nspherical = 3;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 96 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);
    double* PRAGMA_RESTRICT phi_x_tmp = (double*)ALIGNED_MALLOC(64, 96 * sizeof(double));
    ASSUME_ALIGNED(phi_x_tmp, 64);
    double* PRAGMA_RESTRICT phi_y_tmp = (double*)ALIGNED_MALLOC(64, 96 * sizeof(double));
    ASSUME_ALIGNED(phi_y_tmp, 64);
    double* PRAGMA_RESTRICT phi_z_tmp = (double*)ALIGNED_MALLOC(64, 96 * sizeof(double));
    ASSUME_ALIGNED(phi_z_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Combine blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            // Gaussian derivs (gradients)
            const double SX = S1[i] * xc[i];
            const double SY = S1[i] * yc[i];
            const double SZ = S1[i] * zc[i];

            // Density AM=1 Component=X
            phi_tmp[i] = S0[i] * xc[i];

            // Gradient AM=1 Component=X
            phi_x_tmp[i] = SX * xc[i];
            phi_y_tmp[i] = SY * xc[i];
            phi_z_tmp[i] = SZ * xc[i];
            phi_x_tmp[i] += S0[i];

            // Density AM=1 Component=Y
            phi_tmp[32 + i] = S0[i] * yc[i];

            // Gradient AM=1 Component=Y
            phi_x_tmp[32 + i] = SX * yc[i];
            phi_y_tmp[32 + i] = SY * yc[i];
            phi_z_tmp[32 + i] = SZ * yc[i];
            phi_y_tmp[32 + i] += S0[i];

            // Density AM=1 Component=Z
            phi_tmp[64 + i] = S0[i] * zc[i];

            // Gradient AM=1 Component=Z
            phi_x_tmp[64 + i] = SX * zc[i];
            phi_y_tmp[64 + i] = SY * zc[i];
            phi_z_tmp[64 + i] = SZ * zc[i];
            phi_z_tmp[64 + i] += S0[i];

        }

        // Copy data back into outer temps
        if (spherical) {
            // Phi, transform data to outer temps
            gg_cart_to_spherical_L1(remain, phi_tmp, 32, (phi_out + start), npoints);

            // Gradient, transform data to outer temps
            gg_cart_to_spherical_L1(remain, phi_x_tmp, 32, (phi_x_out + start), npoints);
            gg_cart_to_spherical_L1(remain, phi_y_tmp, 32, (phi_y_out + start), npoints);
            gg_cart_to_spherical_L1(remain, phi_z_tmp, 32, (phi_z_out + start), npoints);
            } else {

            // Phi, copy data to outer temps
            block_copy(nout, remain, phi_tmp, 32, (phi_out + start), npoints, 0);

            // Gradient, copy data to outer temps
            block_copy(nout, remain, phi_x_tmp, 32, (phi_x_out + start), npoints, 0);
            block_copy(nout, remain, phi_y_tmp, 32, (phi_y_out + start), npoints, 0);
            block_copy(nout, remain, phi_z_tmp, 32, (phi_z_out + start), npoints, 0);

        }
    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);
    ALIGNED_FREE(phi_x_tmp);
    ALIGNED_FREE(phi_y_tmp);
    ALIGNED_FREE(phi_z_tmp);

}

void gg_collocation_L2_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 6;
    const unsigned long nspherical = 5;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 192 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);
    double* PRAGMA_RESTRICT phi_x_tmp = (double*)ALIGNED_MALLOC(64, 192 * sizeof(double));
    ASSUME_ALIGNED(phi_x_tmp, 64);
    double* PRAGMA_RESTRICT phi_y_tmp = (double*)ALIGNED_MALLOC(64, 192 * sizeof(double));
    ASSUME_ALIGNED(phi_y_tmp, 64);
    double* PRAGMA_RESTRICT phi_z_tmp = (double*)ALIGNED_MALLOC(64, 192 * sizeof(double));
    ASSUME_ALIGNED(phi_z_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Combine blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            // Gaussian derivs (gradients)
            const double SX = S1[i] * xc[i];
            const double SY = S1[i] * yc[i];
            const double SZ = S1[i] * zc[i];

            // Cartesian derivs
            const double xc_pow2 = xc[i] * xc[i];
            const double yc_pow2 = yc[i] * yc[i];
            const double zc_pow2 = zc[i] * zc[i];


            // Density AM=2 Component=XX
            phi_tmp[i] = S0[i] * xc_pow2;

            // Gradient AM=2 Component=XX
            phi_x_tmp[i] = SX * xc_pow2;
            phi_y_tmp[i] = SY * xc_pow2;
            phi_z_tmp[i] = SZ * xc_pow2;
            AX = 2.0 * xc[i];
            phi_x_tmp[i] += S0[i] * AX;

            // Density AM=2 Component=XY
            A = xc[i] * yc[i];
            phi_tmp[32 + i] = S0[i] * A;

            // Gradient AM=2 Component=XY
            phi_x_tmp[32 + i] = SX * A;
            phi_y_tmp[32 + i] = SY * A;
            phi_z_tmp[32 + i] = SZ * A;
            phi_x_tmp[32 + i] += S0[i] * yc[i];
            phi_y_tmp[32 + i] += S0[i] * xc[i];

            // Density AM=2 Component=XZ
            A = xc[i] * zc[i];
            phi_tmp[64 + i] = S0[i] * A;

            // Gradient AM=2 Component=XZ
            phi_x_tmp[64 + i] = SX * A;
            phi_y_tmp[64 + i] = SY * A;
            phi_z_tmp[64 + i] = SZ * A;
            phi_x_tmp[64 + i] += S0[i] * zc[i];
            phi_z_tmp[64 + i] += S0[i] * xc[i];

            // Density AM=2 Component=YY
            phi_tmp[96 + i] = S0[i] * yc_pow2;

            // Gradient AM=2 Component=YY
            phi_x_tmp[96 + i] = SX * yc_pow2;
            phi_y_tmp[96 + i] = SY * yc_pow2;
            phi_z_tmp[96 + i] = SZ * yc_pow2;
            AY = 2.0 * yc[i];
            phi_y_tmp[96 + i] += S0[i] * AY;

            // Density AM=2 Component=YZ
            A = yc[i] * zc[i];
            phi_tmp[128 + i] = S0[i] * A;

            // Gradient AM=2 Component=YZ
            phi_x_tmp[128 + i] = SX * A;
            phi_y_tmp[128 + i] = SY * A;
            phi_z_tmp[128 + i] = SZ * A;
            phi_y_tmp[128 + i] += S0[i] * zc[i];
            phi_z_tmp[128 + i] += S0[i] * yc[i];

            // Density AM=2 Component=ZZ
            phi_tmp[160 + i] = S0[i] * zc_pow2;

            // Gradient AM=2 Component=ZZ
            phi_x_tmp[160 + i] = SX * zc_pow2;
            phi_y_tmp[160 + i] = SY * zc_pow2;
            phi_z_tmp[160 + i] = SZ * zc_pow2;
            AZ = 2.0 * zc[i];
            phi_z_tmp[160 + i] += S0[i] * AZ;

        }

        // Copy data back into outer temps
        if (spherical) {
            // Phi, transform data to outer temps
            gg_cart_to_spherical_L2(remain, phi_tmp, 32, (phi_out + start), npoints);

            // Gradient, transform data to outer temps
            gg_cart_to_spherical_L2(remain, phi_x_tmp, 32, (phi_x_out + start), npoints);
            gg_cart_to_spherical_L2(remain, phi_y_tmp, 32, (phi_y_out + start), npoints);
            gg_cart_to_spherical_L2(remain, phi_z_tmp, 32, (phi_z_out + start), npoints);
            } else {

            // Phi, copy data to outer temps
            block_copy(nout, remain, phi_tmp, 32, (phi_out + start), npoints, 0);

            // Gradient, copy data to outer temps
            block_copy(nout, remain, phi_x_tmp, 32, (phi_x_out + start), npoints, 0);
            block_copy(nout, remain, phi_y_tmp, 32, (phi_y_out + start), npoints, 0);
            block_copy(nout, remain, phi_z_tmp, 32, (phi_z_out + start), npoints, 0);

        }
    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);
    ALIGNED_FREE(phi_x_tmp);
    ALIGNED_FREE(phi_y_tmp);
    ALIGNED_FREE(phi_z_tmp);

}

void gg_collocation_L3_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 10;
    const unsigned long nspherical = 7;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate power temporaries
    double* PRAGMA_RESTRICT xc_pow = (double*)ALIGNED_MALLOC(64, 64 * sizeof(double));
    ASSUME_ALIGNED(xc_pow, 64);
    double* PRAGMA_RESTRICT yc_pow = (double*)ALIGNED_MALLOC(64, 64 * sizeof(double));
    ASSUME_ALIGNED(yc_pow, 64);
    double* PRAGMA_RESTRICT zc_pow = (double*)ALIGNED_MALLOC(64, 64 * sizeof(double));
    ASSUME_ALIGNED(zc_pow, 64);

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 320 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Build powers
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            // Cartesian derivs
            xc_pow[i] = xc[i] * xc[i];
            yc_pow[i] = yc[i] * yc[i];
            zc_pow[i] = zc[i] * zc[i];
            xc_pow[32 + i] = xc_pow[i] * xc[i];
            yc_pow[32 + i] = yc_pow[i] * yc[i];
            zc_pow[32 + i] = zc_pow[i] * zc[i];
        }
        // Combine A blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            phi_tmp[i] = xc_pow[32 + i] * S0[i];
            phi_tmp[32 + i] = xc_pow[i] * yc[i] * S0[i];
            phi_tmp[64 + i] = xc_pow[i] * zc[i] * S0[i];
            phi_tmp[96 + i] = xc[i] * yc_pow[i] * S0[i];
            phi_tmp[128 + i] = xc[i] * yc[i] * zc[i] * S0[i];
            phi_tmp[160 + i] = xc[i] * zc_pow[i] * S0[i];
            phi_tmp[192 + i] = yc_pow[32 + i] * S0[i];
            phi_tmp[224 + i] = yc_pow[i] * zc[i] * S0[i];
            phi_tmp[256 + i] = yc[i] * zc_pow[i] * S0[i];
            phi_tmp[288 + i] = zc_pow[32 + i] * S0[i];
        }

        if (spherical) {
            gg_cart_to_spherical_L3(remain, phi_tmp, 32, (phi_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_out + start), npoints, 0);
        }

        // Combine X blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SX = S1[i] * xc[i];

            phi_tmp[i] = xc_pow[32 + i] * SX;
            phi_tmp[i] += 3.0 * xc_pow[i] * S0[i];

            phi_tmp[32 + i] = xc_pow[i] * yc[i] * SX;
            phi_tmp[32 + i] += 2.0 * xc[i] * yc[i] * S0[i];

            phi_tmp[64 + i] = xc_pow[i] * zc[i] * SX;
            phi_tmp[64 + i] += 2.0 * xc[i] * zc[i] * S0[i];

            phi_tmp[96 + i] = xc[i] * yc_pow[i] * SX;
            phi_tmp[96 + i] += yc_pow[i] * S0[i];

            phi_tmp[128 + i] = xc[i] * yc[i] * zc[i] * SX;
            phi_tmp[128 + i] += yc[i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc[i] * zc_pow[i] * SX;
            phi_tmp[160 + i] += zc_pow[i] * S0[i];

            phi_tmp[192 + i] = yc_pow[32 + i] * SX;

            phi_tmp[224 + i] = yc_pow[i] * zc[i] * SX;

            phi_tmp[256 + i] = yc[i] * zc_pow[i] * SX;

            phi_tmp[288 + i] = zc_pow[32 + i] * SX;

        }

        if (spherical) {
            gg_cart_to_spherical_L3(remain, phi_tmp, 32, (phi_x_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_x_out + start), npoints, 0);
        }

        // Combine Y blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SY = S1[i] * yc[i];

            phi_tmp[i] = xc_pow[32 + i] * SY;

            phi_tmp[32 + i] = xc_pow[i] * yc[i] * SY;
            phi_tmp[32 + i] += xc_pow[i] * S0[i];

            phi_tmp[64 + i] = xc_pow[i] * zc[i] * SY;

            phi_tmp[96 + i] = xc[i] * yc_pow[i] * SY;
            phi_tmp[96 + i] += 2.0 * xc[i] * yc[i] * S0[i];

            phi_tmp[128 + i] = xc[i] * yc[i] * zc[i] * SY;
            phi_tmp[128 + i] += xc[i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc[i] * zc_pow[i] * SY;

            phi_tmp[192 + i] = yc_pow[32 + i] * SY;
            phi_tmp[192 + i] += 3.0 * yc_pow[i] * S0[i];

            phi_tmp[224 + i] = yc_pow[i] * zc[i] * SY;
            phi_tmp[224 + i] += 2.0 * yc[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = yc[i] * zc_pow[i] * SY;
            phi_tmp[256 + i] += zc_pow[i] * S0[i];

            phi_tmp[288 + i] = zc_pow[32 + i] * SY;

        }

        if (spherical) {
            gg_cart_to_spherical_L3(remain, phi_tmp, 32, (phi_y_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_y_out + start), npoints, 0);
        }

        // Combine Z blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SZ = S1[i] * zc[i];

            phi_tmp[i] = xc_pow[32 + i] * SZ;

            phi_tmp[32 + i] = xc_pow[i] * yc[i] * SZ;

            phi_tmp[64 + i] = xc_pow[i] * zc[i] * SZ;
            phi_tmp[64 + i] += xc_pow[i] * S0[i];

            phi_tmp[96 + i] = xc[i] * yc_pow[i] * SZ;

            phi_tmp[128 + i] = xc[i] * yc[i] * zc[i] * SZ;
            phi_tmp[128 + i] += xc[i] * yc[i] * S0[i];

            phi_tmp[160 + i] = xc[i] * zc_pow[i] * SZ;
            phi_tmp[160 + i] += 2.0 * xc[i] * zc[i] * S0[i];

            phi_tmp[192 + i] = yc_pow[32 + i] * SZ;

            phi_tmp[224 + i] = yc_pow[i] * zc[i] * SZ;
            phi_tmp[224 + i] += yc_pow[i] * S0[i];

            phi_tmp[256 + i] = yc[i] * zc_pow[i] * SZ;
            phi_tmp[256 + i] += 2.0 * yc[i] * zc[i] * S0[i];

            phi_tmp[288 + i] = zc_pow[32 + i] * SZ;
            phi_tmp[288 + i] += 3.0 * zc_pow[i] * S0[i];

        }

        if (spherical) {
            gg_cart_to_spherical_L3(remain, phi_tmp, 32, (phi_z_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_z_out + start), npoints, 0);
        }

    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free Power temporaries
    ALIGNED_FREE(xc_pow);
    ALIGNED_FREE(yc_pow);
    ALIGNED_FREE(zc_pow);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);

}

void gg_collocation_L4_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 15;
    const unsigned long nspherical = 9;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate power temporaries
    double* PRAGMA_RESTRICT xc_pow = (double*)ALIGNED_MALLOC(64, 96 * sizeof(double));
    ASSUME_ALIGNED(xc_pow, 64);
    double* PRAGMA_RESTRICT yc_pow = (double*)ALIGNED_MALLOC(64, 96 * sizeof(double));
    ASSUME_ALIGNED(yc_pow, 64);
    double* PRAGMA_RESTRICT zc_pow = (double*)ALIGNED_MALLOC(64, 96 * sizeof(double));
    ASSUME_ALIGNED(zc_pow, 64);

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 480 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Build powers
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            // Cartesian derivs
            xc_pow[i] = xc[i] * xc[i];
            yc_pow[i] = yc[i] * yc[i];
            zc_pow[i] = zc[i] * zc[i];
            xc_pow[32 + i] = xc_pow[i] * xc[i];
            yc_pow[32 + i] = yc_pow[i] * yc[i];
            zc_pow[32 + i] = zc_pow[i] * zc[i];
            xc_pow[64 + i] = xc_pow[32 + i] * xc[i];
            yc_pow[64 + i] = yc_pow[32 + i] * yc[i];
            zc_pow[64 + i] = zc_pow[32 + i] * zc[i];
        }
        // Combine A blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            phi_tmp[i] = xc_pow[64 + i] * S0[i];
            phi_tmp[32 + i] = xc_pow[32 + i] * yc[i] * S0[i];
            phi_tmp[64 + i] = xc_pow[32 + i] * zc[i] * S0[i];
            phi_tmp[96 + i] = xc_pow[i] * yc_pow[i] * S0[i];
            phi_tmp[128 + i] = xc_pow[i] * yc[i] * zc[i] * S0[i];
            phi_tmp[160 + i] = xc_pow[i] * zc_pow[i] * S0[i];
            phi_tmp[192 + i] = xc[i] * yc_pow[32 + i] * S0[i];
            phi_tmp[224 + i] = xc[i] * yc_pow[i] * zc[i] * S0[i];
            phi_tmp[256 + i] = xc[i] * yc[i] * zc_pow[i] * S0[i];
            phi_tmp[288 + i] = xc[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[320 + i] = yc_pow[64 + i] * S0[i];
            phi_tmp[352 + i] = yc_pow[32 + i] * zc[i] * S0[i];
            phi_tmp[384 + i] = yc_pow[i] * zc_pow[i] * S0[i];
            phi_tmp[416 + i] = yc[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[448 + i] = zc_pow[64 + i] * S0[i];
        }

        if (spherical) {
            gg_cart_to_spherical_L4(remain, phi_tmp, 32, (phi_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_out + start), npoints, 0);
        }

        // Combine X blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SX = S1[i] * xc[i];

            phi_tmp[i] = xc_pow[64 + i] * SX;
            phi_tmp[i] += 4.0 * xc_pow[32 + i] * S0[i];

            phi_tmp[32 + i] = xc_pow[32 + i] * yc[i] * SX;
            phi_tmp[32 + i] += 3.0 * xc_pow[i] * yc[i] * S0[i];

            phi_tmp[64 + i] = xc_pow[32 + i] * zc[i] * SX;
            phi_tmp[64 + i] += 3.0 * xc_pow[i] * zc[i] * S0[i];

            phi_tmp[96 + i] = xc_pow[i] * yc_pow[i] * SX;
            phi_tmp[96 + i] += 2.0 * xc[i] * yc_pow[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[i] * yc[i] * zc[i] * SX;
            phi_tmp[128 + i] += 2.0 * xc[i] * yc[i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[i] * zc_pow[i] * SX;
            phi_tmp[160 + i] += 2.0 * xc[i] * zc_pow[i] * S0[i];

            phi_tmp[192 + i] = xc[i] * yc_pow[32 + i] * SX;
            phi_tmp[192 + i] += yc_pow[32 + i] * S0[i];

            phi_tmp[224 + i] = xc[i] * yc_pow[i] * zc[i] * SX;
            phi_tmp[224 + i] += yc_pow[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc[i] * yc[i] * zc_pow[i] * SX;
            phi_tmp[256 + i] += yc[i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc[i] * zc_pow[32 + i] * SX;
            phi_tmp[288 + i] += zc_pow[32 + i] * S0[i];

            phi_tmp[320 + i] = yc_pow[64 + i] * SX;

            phi_tmp[352 + i] = yc_pow[32 + i] * zc[i] * SX;

            phi_tmp[384 + i] = yc_pow[i] * zc_pow[i] * SX;

            phi_tmp[416 + i] = yc[i] * zc_pow[32 + i] * SX;

            phi_tmp[448 + i] = zc_pow[64 + i] * SX;

        }

        if (spherical) {
            gg_cart_to_spherical_L4(remain, phi_tmp, 32, (phi_x_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_x_out + start), npoints, 0);
        }

        // Combine Y blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SY = S1[i] * yc[i];

            phi_tmp[i] = xc_pow[64 + i] * SY;

            phi_tmp[32 + i] = xc_pow[32 + i] * yc[i] * SY;
            phi_tmp[32 + i] += xc_pow[32 + i] * S0[i];

            phi_tmp[64 + i] = xc_pow[32 + i] * zc[i] * SY;

            phi_tmp[96 + i] = xc_pow[i] * yc_pow[i] * SY;
            phi_tmp[96 + i] += 2.0 * xc_pow[i] * yc[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[i] * yc[i] * zc[i] * SY;
            phi_tmp[128 + i] += xc_pow[i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[i] * zc_pow[i] * SY;

            phi_tmp[192 + i] = xc[i] * yc_pow[32 + i] * SY;
            phi_tmp[192 + i] += 3.0 * xc[i] * yc_pow[i] * S0[i];

            phi_tmp[224 + i] = xc[i] * yc_pow[i] * zc[i] * SY;
            phi_tmp[224 + i] += 2.0 * xc[i] * yc[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc[i] * yc[i] * zc_pow[i] * SY;
            phi_tmp[256 + i] += xc[i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc[i] * zc_pow[32 + i] * SY;

            phi_tmp[320 + i] = yc_pow[64 + i] * SY;
            phi_tmp[320 + i] += 4.0 * yc_pow[32 + i] * S0[i];

            phi_tmp[352 + i] = yc_pow[32 + i] * zc[i] * SY;
            phi_tmp[352 + i] += 3.0 * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[384 + i] = yc_pow[i] * zc_pow[i] * SY;
            phi_tmp[384 + i] += 2.0 * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = yc[i] * zc_pow[32 + i] * SY;
            phi_tmp[416 + i] += zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = zc_pow[64 + i] * SY;

        }

        if (spherical) {
            gg_cart_to_spherical_L4(remain, phi_tmp, 32, (phi_y_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_y_out + start), npoints, 0);
        }

        // Combine Z blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SZ = S1[i] * zc[i];

            phi_tmp[i] = xc_pow[64 + i] * SZ;

            phi_tmp[32 + i] = xc_pow[32 + i] * yc[i] * SZ;

            phi_tmp[64 + i] = xc_pow[32 + i] * zc[i] * SZ;
            phi_tmp[64 + i] += xc_pow[32 + i] * S0[i];

            phi_tmp[96 + i] = xc_pow[i] * yc_pow[i] * SZ;

            phi_tmp[128 + i] = xc_pow[i] * yc[i] * zc[i] * SZ;
            phi_tmp[128 + i] += xc_pow[i] * yc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[i] * zc_pow[i] * SZ;
            phi_tmp[160 + i] += 2.0 * xc_pow[i] * zc[i] * S0[i];

            phi_tmp[192 + i] = xc[i] * yc_pow[32 + i] * SZ;

            phi_tmp[224 + i] = xc[i] * yc_pow[i] * zc[i] * SZ;
            phi_tmp[224 + i] += xc[i] * yc_pow[i] * S0[i];

            phi_tmp[256 + i] = xc[i] * yc[i] * zc_pow[i] * SZ;
            phi_tmp[256 + i] += 2.0 * xc[i] * yc[i] * zc[i] * S0[i];

            phi_tmp[288 + i] = xc[i] * zc_pow[32 + i] * SZ;
            phi_tmp[288 + i] += 3.0 * xc[i] * zc_pow[i] * S0[i];

            phi_tmp[320 + i] = yc_pow[64 + i] * SZ;

            phi_tmp[352 + i] = yc_pow[32 + i] * zc[i] * SZ;
            phi_tmp[352 + i] += yc_pow[32 + i] * S0[i];

            phi_tmp[384 + i] = yc_pow[i] * zc_pow[i] * SZ;
            phi_tmp[384 + i] += 2.0 * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[416 + i] = yc[i] * zc_pow[32 + i] * SZ;
            phi_tmp[416 + i] += 3.0 * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[448 + i] = zc_pow[64 + i] * SZ;
            phi_tmp[448 + i] += 4.0 * zc_pow[32 + i] * S0[i];

        }

        if (spherical) {
            gg_cart_to_spherical_L4(remain, phi_tmp, 32, (phi_z_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_z_out + start), npoints, 0);
        }

    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free Power temporaries
    ALIGNED_FREE(xc_pow);
    ALIGNED_FREE(yc_pow);
    ALIGNED_FREE(zc_pow);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);

}

void gg_collocation_L5_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 21;
    const unsigned long nspherical = 11;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate power temporaries
    double* PRAGMA_RESTRICT xc_pow = (double*)ALIGNED_MALLOC(64, 128 * sizeof(double));
    ASSUME_ALIGNED(xc_pow, 64);
    double* PRAGMA_RESTRICT yc_pow = (double*)ALIGNED_MALLOC(64, 128 * sizeof(double));
    ASSUME_ALIGNED(yc_pow, 64);
    double* PRAGMA_RESTRICT zc_pow = (double*)ALIGNED_MALLOC(64, 128 * sizeof(double));
    ASSUME_ALIGNED(zc_pow, 64);

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 672 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Build powers
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            // Cartesian derivs
            xc_pow[i] = xc[i] * xc[i];
            yc_pow[i] = yc[i] * yc[i];
            zc_pow[i] = zc[i] * zc[i];
            xc_pow[32 + i] = xc_pow[i] * xc[i];
            yc_pow[32 + i] = yc_pow[i] * yc[i];
            zc_pow[32 + i] = zc_pow[i] * zc[i];
            xc_pow[64 + i] = xc_pow[32 + i] * xc[i];
            yc_pow[64 + i] = yc_pow[32 + i] * yc[i];
            zc_pow[64 + i] = zc_pow[32 + i] * zc[i];
            xc_pow[96 + i] = xc_pow[64 + i] * xc[i];
            yc_pow[96 + i] = yc_pow[64 + i] * yc[i];
            zc_pow[96 + i] = zc_pow[64 + i] * zc[i];
        }
        // Combine A blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            phi_tmp[i] = xc_pow[96 + i] * S0[i];
            phi_tmp[32 + i] = xc_pow[64 + i] * yc[i] * S0[i];
            phi_tmp[64 + i] = xc_pow[64 + i] * zc[i] * S0[i];
            phi_tmp[96 + i] = xc_pow[32 + i] * yc_pow[i] * S0[i];
            phi_tmp[128 + i] = xc_pow[32 + i] * yc[i] * zc[i] * S0[i];
            phi_tmp[160 + i] = xc_pow[32 + i] * zc_pow[i] * S0[i];
            phi_tmp[192 + i] = xc_pow[i] * yc_pow[32 + i] * S0[i];
            phi_tmp[224 + i] = xc_pow[i] * yc_pow[i] * zc[i] * S0[i];
            phi_tmp[256 + i] = xc_pow[i] * yc[i] * zc_pow[i] * S0[i];
            phi_tmp[288 + i] = xc_pow[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[320 + i] = xc[i] * yc_pow[64 + i] * S0[i];
            phi_tmp[352 + i] = xc[i] * yc_pow[32 + i] * zc[i] * S0[i];
            phi_tmp[384 + i] = xc[i] * yc_pow[i] * zc_pow[i] * S0[i];
            phi_tmp[416 + i] = xc[i] * yc[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[448 + i] = xc[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[480 + i] = yc_pow[96 + i] * S0[i];
            phi_tmp[512 + i] = yc_pow[64 + i] * zc[i] * S0[i];
            phi_tmp[544 + i] = yc_pow[32 + i] * zc_pow[i] * S0[i];
            phi_tmp[576 + i] = yc_pow[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[608 + i] = yc[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[640 + i] = zc_pow[96 + i] * S0[i];
        }

        if (spherical) {
            gg_cart_to_spherical_L5(remain, phi_tmp, 32, (phi_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_out + start), npoints, 0);
        }

        // Combine X blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SX = S1[i] * xc[i];

            phi_tmp[i] = xc_pow[96 + i] * SX;
            phi_tmp[i] += 5.0 * xc_pow[64 + i] * S0[i];

            phi_tmp[32 + i] = xc_pow[64 + i] * yc[i] * SX;
            phi_tmp[32 + i] += 4.0 * xc_pow[32 + i] * yc[i] * S0[i];

            phi_tmp[64 + i] = xc_pow[64 + i] * zc[i] * SX;
            phi_tmp[64 + i] += 4.0 * xc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[96 + i] = xc_pow[32 + i] * yc_pow[i] * SX;
            phi_tmp[96 + i] += 3.0 * xc_pow[i] * yc_pow[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[32 + i] * yc[i] * zc[i] * SX;
            phi_tmp[128 + i] += 3.0 * xc_pow[i] * yc[i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[32 + i] * zc_pow[i] * SX;
            phi_tmp[160 + i] += 3.0 * xc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[192 + i] = xc_pow[i] * yc_pow[32 + i] * SX;
            phi_tmp[192 + i] += 2.0 * xc[i] * yc_pow[32 + i] * S0[i];

            phi_tmp[224 + i] = xc_pow[i] * yc_pow[i] * zc[i] * SX;
            phi_tmp[224 + i] += 2.0 * xc[i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[i] * yc[i] * zc_pow[i] * SX;
            phi_tmp[256 + i] += 2.0 * xc[i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[i] * zc_pow[32 + i] * SX;
            phi_tmp[288 + i] += 2.0 * xc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[320 + i] = xc[i] * yc_pow[64 + i] * SX;
            phi_tmp[320 + i] += yc_pow[64 + i] * S0[i];

            phi_tmp[352 + i] = xc[i] * yc_pow[32 + i] * zc[i] * SX;
            phi_tmp[352 + i] += yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[384 + i] = xc[i] * yc_pow[i] * zc_pow[i] * SX;
            phi_tmp[384 + i] += yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = xc[i] * yc[i] * zc_pow[32 + i] * SX;
            phi_tmp[416 + i] += yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = xc[i] * zc_pow[64 + i] * SX;
            phi_tmp[448 + i] += zc_pow[64 + i] * S0[i];

            phi_tmp[480 + i] = yc_pow[96 + i] * SX;

            phi_tmp[512 + i] = yc_pow[64 + i] * zc[i] * SX;

            phi_tmp[544 + i] = yc_pow[32 + i] * zc_pow[i] * SX;

            phi_tmp[576 + i] = yc_pow[i] * zc_pow[32 + i] * SX;

            phi_tmp[608 + i] = yc[i] * zc_pow[64 + i] * SX;

            phi_tmp[640 + i] = zc_pow[96 + i] * SX;

        }

        if (spherical) {
            gg_cart_to_spherical_L5(remain, phi_tmp, 32, (phi_x_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_x_out + start), npoints, 0);
        }

        // Combine Y blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SY = S1[i] * yc[i];

            phi_tmp[i] = xc_pow[96 + i] * SY;

            phi_tmp[32 + i] = xc_pow[64 + i] * yc[i] * SY;
            phi_tmp[32 + i] += xc_pow[64 + i] * S0[i];

            phi_tmp[64 + i] = xc_pow[64 + i] * zc[i] * SY;

            phi_tmp[96 + i] = xc_pow[32 + i] * yc_pow[i] * SY;
            phi_tmp[96 + i] += 2.0 * xc_pow[32 + i] * yc[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[32 + i] * yc[i] * zc[i] * SY;
            phi_tmp[128 + i] += xc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[32 + i] * zc_pow[i] * SY;

            phi_tmp[192 + i] = xc_pow[i] * yc_pow[32 + i] * SY;
            phi_tmp[192 + i] += 3.0 * xc_pow[i] * yc_pow[i] * S0[i];

            phi_tmp[224 + i] = xc_pow[i] * yc_pow[i] * zc[i] * SY;
            phi_tmp[224 + i] += 2.0 * xc_pow[i] * yc[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[i] * yc[i] * zc_pow[i] * SY;
            phi_tmp[256 + i] += xc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[i] * zc_pow[32 + i] * SY;

            phi_tmp[320 + i] = xc[i] * yc_pow[64 + i] * SY;
            phi_tmp[320 + i] += 4.0 * xc[i] * yc_pow[32 + i] * S0[i];

            phi_tmp[352 + i] = xc[i] * yc_pow[32 + i] * zc[i] * SY;
            phi_tmp[352 + i] += 3.0 * xc[i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[384 + i] = xc[i] * yc_pow[i] * zc_pow[i] * SY;
            phi_tmp[384 + i] += 2.0 * xc[i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = xc[i] * yc[i] * zc_pow[32 + i] * SY;
            phi_tmp[416 + i] += xc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = xc[i] * zc_pow[64 + i] * SY;

            phi_tmp[480 + i] = yc_pow[96 + i] * SY;
            phi_tmp[480 + i] += 5.0 * yc_pow[64 + i] * S0[i];

            phi_tmp[512 + i] = yc_pow[64 + i] * zc[i] * SY;
            phi_tmp[512 + i] += 4.0 * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[544 + i] = yc_pow[32 + i] * zc_pow[i] * SY;
            phi_tmp[544 + i] += 3.0 * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[576 + i] = yc_pow[i] * zc_pow[32 + i] * SY;
            phi_tmp[576 + i] += 2.0 * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[608 + i] = yc[i] * zc_pow[64 + i] * SY;
            phi_tmp[608 + i] += zc_pow[64 + i] * S0[i];

            phi_tmp[640 + i] = zc_pow[96 + i] * SY;

        }

        if (spherical) {
            gg_cart_to_spherical_L5(remain, phi_tmp, 32, (phi_y_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_y_out + start), npoints, 0);
        }

        // Combine Z blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SZ = S1[i] * zc[i];

            phi_tmp[i] = xc_pow[96 + i] * SZ;

            phi_tmp[32 + i] = xc_pow[64 + i] * yc[i] * SZ;

            phi_tmp[64 + i] = xc_pow[64 + i] * zc[i] * SZ;
            phi_tmp[64 + i] += xc_pow[64 + i] * S0[i];

            phi_tmp[96 + i] = xc_pow[32 + i] * yc_pow[i] * SZ;

            phi_tmp[128 + i] = xc_pow[32 + i] * yc[i] * zc[i] * SZ;
            phi_tmp[128 + i] += xc_pow[32 + i] * yc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[32 + i] * zc_pow[i] * SZ;
            phi_tmp[160 + i] += 2.0 * xc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[192 + i] = xc_pow[i] * yc_pow[32 + i] * SZ;

            phi_tmp[224 + i] = xc_pow[i] * yc_pow[i] * zc[i] * SZ;
            phi_tmp[224 + i] += xc_pow[i] * yc_pow[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[i] * yc[i] * zc_pow[i] * SZ;
            phi_tmp[256 + i] += 2.0 * xc_pow[i] * yc[i] * zc[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[i] * zc_pow[32 + i] * SZ;
            phi_tmp[288 + i] += 3.0 * xc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[320 + i] = xc[i] * yc_pow[64 + i] * SZ;

            phi_tmp[352 + i] = xc[i] * yc_pow[32 + i] * zc[i] * SZ;
            phi_tmp[352 + i] += xc[i] * yc_pow[32 + i] * S0[i];

            phi_tmp[384 + i] = xc[i] * yc_pow[i] * zc_pow[i] * SZ;
            phi_tmp[384 + i] += 2.0 * xc[i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[416 + i] = xc[i] * yc[i] * zc_pow[32 + i] * SZ;
            phi_tmp[416 + i] += 3.0 * xc[i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[448 + i] = xc[i] * zc_pow[64 + i] * SZ;
            phi_tmp[448 + i] += 4.0 * xc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[480 + i] = yc_pow[96 + i] * SZ;

            phi_tmp[512 + i] = yc_pow[64 + i] * zc[i] * SZ;
            phi_tmp[512 + i] += yc_pow[64 + i] * S0[i];

            phi_tmp[544 + i] = yc_pow[32 + i] * zc_pow[i] * SZ;
            phi_tmp[544 + i] += 2.0 * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[576 + i] = yc_pow[i] * zc_pow[32 + i] * SZ;
            phi_tmp[576 + i] += 3.0 * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[608 + i] = yc[i] * zc_pow[64 + i] * SZ;
            phi_tmp[608 + i] += 4.0 * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[640 + i] = zc_pow[96 + i] * SZ;
            phi_tmp[640 + i] += 5.0 * zc_pow[64 + i] * S0[i];

        }

        if (spherical) {
            gg_cart_to_spherical_L5(remain, phi_tmp, 32, (phi_z_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_z_out + start), npoints, 0);
        }

    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free Power temporaries
    ALIGNED_FREE(xc_pow);
    ALIGNED_FREE(yc_pow);
    ALIGNED_FREE(zc_pow);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);

}

void gg_collocation_L6_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 28;
    const unsigned long nspherical = 13;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate power temporaries
    double* PRAGMA_RESTRICT xc_pow = (double*)ALIGNED_MALLOC(64, 160 * sizeof(double));
    ASSUME_ALIGNED(xc_pow, 64);
    double* PRAGMA_RESTRICT yc_pow = (double*)ALIGNED_MALLOC(64, 160 * sizeof(double));
    ASSUME_ALIGNED(yc_pow, 64);
    double* PRAGMA_RESTRICT zc_pow = (double*)ALIGNED_MALLOC(64, 160 * sizeof(double));
    ASSUME_ALIGNED(zc_pow, 64);

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 896 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Build powers
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            // Cartesian derivs
            xc_pow[i] = xc[i] * xc[i];
            yc_pow[i] = yc[i] * yc[i];
            zc_pow[i] = zc[i] * zc[i];
            xc_pow[32 + i] = xc_pow[i] * xc[i];
            yc_pow[32 + i] = yc_pow[i] * yc[i];
            zc_pow[32 + i] = zc_pow[i] * zc[i];
            xc_pow[64 + i] = xc_pow[32 + i] * xc[i];
            yc_pow[64 + i] = yc_pow[32 + i] * yc[i];
            zc_pow[64 + i] = zc_pow[32 + i] * zc[i];
            xc_pow[96 + i] = xc_pow[64 + i] * xc[i];
            yc_pow[96 + i] = yc_pow[64 + i] * yc[i];
            zc_pow[96 + i] = zc_pow[64 + i] * zc[i];
            xc_pow[128 + i] = xc_pow[96 + i] * xc[i];
            yc_pow[128 + i] = yc_pow[96 + i] * yc[i];
            zc_pow[128 + i] = zc_pow[96 + i] * zc[i];
        }
        // Combine A blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            phi_tmp[i] = xc_pow[128 + i] * S0[i];
            phi_tmp[32 + i] = xc_pow[96 + i] * yc[i] * S0[i];
            phi_tmp[64 + i] = xc_pow[96 + i] * zc[i] * S0[i];
            phi_tmp[96 + i] = xc_pow[64 + i] * yc_pow[i] * S0[i];
            phi_tmp[128 + i] = xc_pow[64 + i] * yc[i] * zc[i] * S0[i];
            phi_tmp[160 + i] = xc_pow[64 + i] * zc_pow[i] * S0[i];
            phi_tmp[192 + i] = xc_pow[32 + i] * yc_pow[32 + i] * S0[i];
            phi_tmp[224 + i] = xc_pow[32 + i] * yc_pow[i] * zc[i] * S0[i];
            phi_tmp[256 + i] = xc_pow[32 + i] * yc[i] * zc_pow[i] * S0[i];
            phi_tmp[288 + i] = xc_pow[32 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[320 + i] = xc_pow[i] * yc_pow[64 + i] * S0[i];
            phi_tmp[352 + i] = xc_pow[i] * yc_pow[32 + i] * zc[i] * S0[i];
            phi_tmp[384 + i] = xc_pow[i] * yc_pow[i] * zc_pow[i] * S0[i];
            phi_tmp[416 + i] = xc_pow[i] * yc[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[448 + i] = xc_pow[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[480 + i] = xc[i] * yc_pow[96 + i] * S0[i];
            phi_tmp[512 + i] = xc[i] * yc_pow[64 + i] * zc[i] * S0[i];
            phi_tmp[544 + i] = xc[i] * yc_pow[32 + i] * zc_pow[i] * S0[i];
            phi_tmp[576 + i] = xc[i] * yc_pow[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[608 + i] = xc[i] * yc[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[640 + i] = xc[i] * zc_pow[96 + i] * S0[i];
            phi_tmp[672 + i] = yc_pow[128 + i] * S0[i];
            phi_tmp[704 + i] = yc_pow[96 + i] * zc[i] * S0[i];
            phi_tmp[736 + i] = yc_pow[64 + i] * zc_pow[i] * S0[i];
            phi_tmp[768 + i] = yc_pow[32 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[800 + i] = yc_pow[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[832 + i] = yc[i] * zc_pow[96 + i] * S0[i];
            phi_tmp[864 + i] = zc_pow[128 + i] * S0[i];
        }

        if (spherical) {
            gg_cart_to_spherical_L6(remain, phi_tmp, 32, (phi_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_out + start), npoints, 0);
        }

        // Combine X blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SX = S1[i] * xc[i];

            phi_tmp[i] = xc_pow[128 + i] * SX;
            phi_tmp[i] += 6.0 * xc_pow[96 + i] * S0[i];

            phi_tmp[32 + i] = xc_pow[96 + i] * yc[i] * SX;
            phi_tmp[32 + i] += 5.0 * xc_pow[64 + i] * yc[i] * S0[i];

            phi_tmp[64 + i] = xc_pow[96 + i] * zc[i] * SX;
            phi_tmp[64 + i] += 5.0 * xc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[96 + i] = xc_pow[64 + i] * yc_pow[i] * SX;
            phi_tmp[96 + i] += 4.0 * xc_pow[32 + i] * yc_pow[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[64 + i] * yc[i] * zc[i] * SX;
            phi_tmp[128 + i] += 4.0 * xc_pow[32 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[64 + i] * zc_pow[i] * SX;
            phi_tmp[160 + i] += 4.0 * xc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[192 + i] = xc_pow[32 + i] * yc_pow[32 + i] * SX;
            phi_tmp[192 + i] += 3.0 * xc_pow[i] * yc_pow[32 + i] * S0[i];

            phi_tmp[224 + i] = xc_pow[32 + i] * yc_pow[i] * zc[i] * SX;
            phi_tmp[224 + i] += 3.0 * xc_pow[i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[32 + i] * yc[i] * zc_pow[i] * SX;
            phi_tmp[256 + i] += 3.0 * xc_pow[i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[32 + i] * zc_pow[32 + i] * SX;
            phi_tmp[288 + i] += 3.0 * xc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[320 + i] = xc_pow[i] * yc_pow[64 + i] * SX;
            phi_tmp[320 + i] += 2.0 * xc[i] * yc_pow[64 + i] * S0[i];

            phi_tmp[352 + i] = xc_pow[i] * yc_pow[32 + i] * zc[i] * SX;
            phi_tmp[352 + i] += 2.0 * xc[i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[384 + i] = xc_pow[i] * yc_pow[i] * zc_pow[i] * SX;
            phi_tmp[384 + i] += 2.0 * xc[i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[i] * yc[i] * zc_pow[32 + i] * SX;
            phi_tmp[416 + i] += 2.0 * xc[i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = xc_pow[i] * zc_pow[64 + i] * SX;
            phi_tmp[448 + i] += 2.0 * xc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[480 + i] = xc[i] * yc_pow[96 + i] * SX;
            phi_tmp[480 + i] += yc_pow[96 + i] * S0[i];

            phi_tmp[512 + i] = xc[i] * yc_pow[64 + i] * zc[i] * SX;
            phi_tmp[512 + i] += yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[544 + i] = xc[i] * yc_pow[32 + i] * zc_pow[i] * SX;
            phi_tmp[544 + i] += yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[576 + i] = xc[i] * yc_pow[i] * zc_pow[32 + i] * SX;
            phi_tmp[576 + i] += yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[608 + i] = xc[i] * yc[i] * zc_pow[64 + i] * SX;
            phi_tmp[608 + i] += yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[640 + i] = xc[i] * zc_pow[96 + i] * SX;
            phi_tmp[640 + i] += zc_pow[96 + i] * S0[i];

            phi_tmp[672 + i] = yc_pow[128 + i] * SX;

            phi_tmp[704 + i] = yc_pow[96 + i] * zc[i] * SX;

            phi_tmp[736 + i] = yc_pow[64 + i] * zc_pow[i] * SX;

            phi_tmp[768 + i] = yc_pow[32 + i] * zc_pow[32 + i] * SX;

            phi_tmp[800 + i] = yc_pow[i] * zc_pow[64 + i] * SX;

            phi_tmp[832 + i] = yc[i] * zc_pow[96 + i] * SX;

            phi_tmp[864 + i] = zc_pow[128 + i] * SX;

        }

        if (spherical) {
            gg_cart_to_spherical_L6(remain, phi_tmp, 32, (phi_x_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_x_out + start), npoints, 0);
        }

        // Combine Y blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SY = S1[i] * yc[i];

            phi_tmp[i] = xc_pow[128 + i] * SY;

            phi_tmp[32 + i] = xc_pow[96 + i] * yc[i] * SY;
            phi_tmp[32 + i] += xc_pow[96 + i] * S0[i];

            phi_tmp[64 + i] = xc_pow[96 + i] * zc[i] * SY;

            phi_tmp[96 + i] = xc_pow[64 + i] * yc_pow[i] * SY;
            phi_tmp[96 + i] += 2.0 * xc_pow[64 + i] * yc[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[64 + i] * yc[i] * zc[i] * SY;
            phi_tmp[128 + i] += xc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[64 + i] * zc_pow[i] * SY;

            phi_tmp[192 + i] = xc_pow[32 + i] * yc_pow[32 + i] * SY;
            phi_tmp[192 + i] += 3.0 * xc_pow[32 + i] * yc_pow[i] * S0[i];

            phi_tmp[224 + i] = xc_pow[32 + i] * yc_pow[i] * zc[i] * SY;
            phi_tmp[224 + i] += 2.0 * xc_pow[32 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[32 + i] * yc[i] * zc_pow[i] * SY;
            phi_tmp[256 + i] += xc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[32 + i] * zc_pow[32 + i] * SY;

            phi_tmp[320 + i] = xc_pow[i] * yc_pow[64 + i] * SY;
            phi_tmp[320 + i] += 4.0 * xc_pow[i] * yc_pow[32 + i] * S0[i];

            phi_tmp[352 + i] = xc_pow[i] * yc_pow[32 + i] * zc[i] * SY;
            phi_tmp[352 + i] += 3.0 * xc_pow[i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[384 + i] = xc_pow[i] * yc_pow[i] * zc_pow[i] * SY;
            phi_tmp[384 + i] += 2.0 * xc_pow[i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[i] * yc[i] * zc_pow[32 + i] * SY;
            phi_tmp[416 + i] += xc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = xc_pow[i] * zc_pow[64 + i] * SY;

            phi_tmp[480 + i] = xc[i] * yc_pow[96 + i] * SY;
            phi_tmp[480 + i] += 5.0 * xc[i] * yc_pow[64 + i] * S0[i];

            phi_tmp[512 + i] = xc[i] * yc_pow[64 + i] * zc[i] * SY;
            phi_tmp[512 + i] += 4.0 * xc[i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[544 + i] = xc[i] * yc_pow[32 + i] * zc_pow[i] * SY;
            phi_tmp[544 + i] += 3.0 * xc[i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[576 + i] = xc[i] * yc_pow[i] * zc_pow[32 + i] * SY;
            phi_tmp[576 + i] += 2.0 * xc[i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[608 + i] = xc[i] * yc[i] * zc_pow[64 + i] * SY;
            phi_tmp[608 + i] += xc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[640 + i] = xc[i] * zc_pow[96 + i] * SY;

            phi_tmp[672 + i] = yc_pow[128 + i] * SY;
            phi_tmp[672 + i] += 6.0 * yc_pow[96 + i] * S0[i];

            phi_tmp[704 + i] = yc_pow[96 + i] * zc[i] * SY;
            phi_tmp[704 + i] += 5.0 * yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[736 + i] = yc_pow[64 + i] * zc_pow[i] * SY;
            phi_tmp[736 + i] += 4.0 * yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[768 + i] = yc_pow[32 + i] * zc_pow[32 + i] * SY;
            phi_tmp[768 + i] += 3.0 * yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[800 + i] = yc_pow[i] * zc_pow[64 + i] * SY;
            phi_tmp[800 + i] += 2.0 * yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[832 + i] = yc[i] * zc_pow[96 + i] * SY;
            phi_tmp[832 + i] += zc_pow[96 + i] * S0[i];

            phi_tmp[864 + i] = zc_pow[128 + i] * SY;

        }

        if (spherical) {
            gg_cart_to_spherical_L6(remain, phi_tmp, 32, (phi_y_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_y_out + start), npoints, 0);
        }

        // Combine Z blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SZ = S1[i] * zc[i];

            phi_tmp[i] = xc_pow[128 + i] * SZ;

            phi_tmp[32 + i] = xc_pow[96 + i] * yc[i] * SZ;

            phi_tmp[64 + i] = xc_pow[96 + i] * zc[i] * SZ;
            phi_tmp[64 + i] += xc_pow[96 + i] * S0[i];

            phi_tmp[96 + i] = xc_pow[64 + i] * yc_pow[i] * SZ;

            phi_tmp[128 + i] = xc_pow[64 + i] * yc[i] * zc[i] * SZ;
            phi_tmp[128 + i] += xc_pow[64 + i] * yc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[64 + i] * zc_pow[i] * SZ;
            phi_tmp[160 + i] += 2.0 * xc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[192 + i] = xc_pow[32 + i] * yc_pow[32 + i] * SZ;

            phi_tmp[224 + i] = xc_pow[32 + i] * yc_pow[i] * zc[i] * SZ;
            phi_tmp[224 + i] += xc_pow[32 + i] * yc_pow[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[32 + i] * yc[i] * zc_pow[i] * SZ;
            phi_tmp[256 + i] += 2.0 * xc_pow[32 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[32 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[288 + i] += 3.0 * xc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[320 + i] = xc_pow[i] * yc_pow[64 + i] * SZ;

            phi_tmp[352 + i] = xc_pow[i] * yc_pow[32 + i] * zc[i] * SZ;
            phi_tmp[352 + i] += xc_pow[i] * yc_pow[32 + i] * S0[i];

            phi_tmp[384 + i] = xc_pow[i] * yc_pow[i] * zc_pow[i] * SZ;
            phi_tmp[384 + i] += 2.0 * xc_pow[i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[i] * yc[i] * zc_pow[32 + i] * SZ;
            phi_tmp[416 + i] += 3.0 * xc_pow[i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[448 + i] = xc_pow[i] * zc_pow[64 + i] * SZ;
            phi_tmp[448 + i] += 4.0 * xc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[480 + i] = xc[i] * yc_pow[96 + i] * SZ;

            phi_tmp[512 + i] = xc[i] * yc_pow[64 + i] * zc[i] * SZ;
            phi_tmp[512 + i] += xc[i] * yc_pow[64 + i] * S0[i];

            phi_tmp[544 + i] = xc[i] * yc_pow[32 + i] * zc_pow[i] * SZ;
            phi_tmp[544 + i] += 2.0 * xc[i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[576 + i] = xc[i] * yc_pow[i] * zc_pow[32 + i] * SZ;
            phi_tmp[576 + i] += 3.0 * xc[i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[608 + i] = xc[i] * yc[i] * zc_pow[64 + i] * SZ;
            phi_tmp[608 + i] += 4.0 * xc[i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[640 + i] = xc[i] * zc_pow[96 + i] * SZ;
            phi_tmp[640 + i] += 5.0 * xc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[672 + i] = yc_pow[128 + i] * SZ;

            phi_tmp[704 + i] = yc_pow[96 + i] * zc[i] * SZ;
            phi_tmp[704 + i] += yc_pow[96 + i] * S0[i];

            phi_tmp[736 + i] = yc_pow[64 + i] * zc_pow[i] * SZ;
            phi_tmp[736 + i] += 2.0 * yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[768 + i] = yc_pow[32 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[768 + i] += 3.0 * yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[800 + i] = yc_pow[i] * zc_pow[64 + i] * SZ;
            phi_tmp[800 + i] += 4.0 * yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[832 + i] = yc[i] * zc_pow[96 + i] * SZ;
            phi_tmp[832 + i] += 5.0 * yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[864 + i] = zc_pow[128 + i] * SZ;
            phi_tmp[864 + i] += 6.0 * zc_pow[96 + i] * S0[i];

        }

        if (spherical) {
            gg_cart_to_spherical_L6(remain, phi_tmp, 32, (phi_z_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_z_out + start), npoints, 0);
        }

    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free Power temporaries
    ALIGNED_FREE(xc_pow);
    ALIGNED_FREE(yc_pow);
    ALIGNED_FREE(zc_pow);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);

}

void gg_collocation_L7_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 36;
    const unsigned long nspherical = 15;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate power temporaries
    double* PRAGMA_RESTRICT xc_pow = (double*)ALIGNED_MALLOC(64, 192 * sizeof(double));
    ASSUME_ALIGNED(xc_pow, 64);
    double* PRAGMA_RESTRICT yc_pow = (double*)ALIGNED_MALLOC(64, 192 * sizeof(double));
    ASSUME_ALIGNED(yc_pow, 64);
    double* PRAGMA_RESTRICT zc_pow = (double*)ALIGNED_MALLOC(64, 192 * sizeof(double));
    ASSUME_ALIGNED(zc_pow, 64);

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 1152 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Build powers
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            // Cartesian derivs
            xc_pow[i] = xc[i] * xc[i];
            yc_pow[i] = yc[i] * yc[i];
            zc_pow[i] = zc[i] * zc[i];
            xc_pow[32 + i] = xc_pow[i] * xc[i];
            yc_pow[32 + i] = yc_pow[i] * yc[i];
            zc_pow[32 + i] = zc_pow[i] * zc[i];
            xc_pow[64 + i] = xc_pow[32 + i] * xc[i];
            yc_pow[64 + i] = yc_pow[32 + i] * yc[i];
            zc_pow[64 + i] = zc_pow[32 + i] * zc[i];
            xc_pow[96 + i] = xc_pow[64 + i] * xc[i];
            yc_pow[96 + i] = yc_pow[64 + i] * yc[i];
            zc_pow[96 + i] = zc_pow[64 + i] * zc[i];
            xc_pow[128 + i] = xc_pow[96 + i] * xc[i];
            yc_pow[128 + i] = yc_pow[96 + i] * yc[i];
            zc_pow[128 + i] = zc_pow[96 + i] * zc[i];
            xc_pow[160 + i] = xc_pow[128 + i] * xc[i];
            yc_pow[160 + i] = yc_pow[128 + i] * yc[i];
            zc_pow[160 + i] = zc_pow[128 + i] * zc[i];
        }
        // Combine A blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            phi_tmp[i] = xc_pow[160 + i] * S0[i];
            phi_tmp[32 + i] = xc_pow[128 + i] * yc[i] * S0[i];
            phi_tmp[64 + i] = xc_pow[128 + i] * zc[i] * S0[i];
            phi_tmp[96 + i] = xc_pow[96 + i] * yc_pow[i] * S0[i];
            phi_tmp[128 + i] = xc_pow[96 + i] * yc[i] * zc[i] * S0[i];
            phi_tmp[160 + i] = xc_pow[96 + i] * zc_pow[i] * S0[i];
            phi_tmp[192 + i] = xc_pow[64 + i] * yc_pow[32 + i] * S0[i];
            phi_tmp[224 + i] = xc_pow[64 + i] * yc_pow[i] * zc[i] * S0[i];
            phi_tmp[256 + i] = xc_pow[64 + i] * yc[i] * zc_pow[i] * S0[i];
            phi_tmp[288 + i] = xc_pow[64 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[320 + i] = xc_pow[32 + i] * yc_pow[64 + i] * S0[i];
            phi_tmp[352 + i] = xc_pow[32 + i] * yc_pow[32 + i] * zc[i] * S0[i];
            phi_tmp[384 + i] = xc_pow[32 + i] * yc_pow[i] * zc_pow[i] * S0[i];
            phi_tmp[416 + i] = xc_pow[32 + i] * yc[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[448 + i] = xc_pow[32 + i] * zc_pow[64 + i] * S0[i];
            phi_tmp[480 + i] = xc_pow[i] * yc_pow[96 + i] * S0[i];
            phi_tmp[512 + i] = xc_pow[i] * yc_pow[64 + i] * zc[i] * S0[i];
            phi_tmp[544 + i] = xc_pow[i] * yc_pow[32 + i] * zc_pow[i] * S0[i];
            phi_tmp[576 + i] = xc_pow[i] * yc_pow[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[608 + i] = xc_pow[i] * yc[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[640 + i] = xc_pow[i] * zc_pow[96 + i] * S0[i];
            phi_tmp[672 + i] = xc[i] * yc_pow[128 + i] * S0[i];
            phi_tmp[704 + i] = xc[i] * yc_pow[96 + i] * zc[i] * S0[i];
            phi_tmp[736 + i] = xc[i] * yc_pow[64 + i] * zc_pow[i] * S0[i];
            phi_tmp[768 + i] = xc[i] * yc_pow[32 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[800 + i] = xc[i] * yc_pow[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[832 + i] = xc[i] * yc[i] * zc_pow[96 + i] * S0[i];
            phi_tmp[864 + i] = xc[i] * zc_pow[128 + i] * S0[i];
            phi_tmp[896 + i] = yc_pow[160 + i] * S0[i];
            phi_tmp[928 + i] = yc_pow[128 + i] * zc[i] * S0[i];
            phi_tmp[960 + i] = yc_pow[96 + i] * zc_pow[i] * S0[i];
            phi_tmp[992 + i] = yc_pow[64 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[1024 + i] = yc_pow[32 + i] * zc_pow[64 + i] * S0[i];
            phi_tmp[1056 + i] = yc_pow[i] * zc_pow[96 + i] * S0[i];
            phi_tmp[1088 + i] = yc[i] * zc_pow[128 + i] * S0[i];
            phi_tmp[1120 + i] = zc_pow[160 + i] * S0[i];
        }

        if (spherical) {
            gg_cart_to_spherical_L7(remain, phi_tmp, 32, (phi_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_out + start), npoints, 0);
        }

        // Combine X blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SX = S1[i] * xc[i];

            phi_tmp[i] = xc_pow[160 + i] * SX;
            phi_tmp[i] += 7.0 * xc_pow[128 + i] * S0[i];

            phi_tmp[32 + i] = xc_pow[128 + i] * yc[i] * SX;
            phi_tmp[32 + i] += 6.0 * xc_pow[96 + i] * yc[i] * S0[i];

            phi_tmp[64 + i] = xc_pow[128 + i] * zc[i] * SX;
            phi_tmp[64 + i] += 6.0 * xc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[96 + i] = xc_pow[96 + i] * yc_pow[i] * SX;
            phi_tmp[96 + i] += 5.0 * xc_pow[64 + i] * yc_pow[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[96 + i] * yc[i] * zc[i] * SX;
            phi_tmp[128 + i] += 5.0 * xc_pow[64 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[96 + i] * zc_pow[i] * SX;
            phi_tmp[160 + i] += 5.0 * xc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[192 + i] = xc_pow[64 + i] * yc_pow[32 + i] * SX;
            phi_tmp[192 + i] += 4.0 * xc_pow[32 + i] * yc_pow[32 + i] * S0[i];

            phi_tmp[224 + i] = xc_pow[64 + i] * yc_pow[i] * zc[i] * SX;
            phi_tmp[224 + i] += 4.0 * xc_pow[32 + i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[64 + i] * yc[i] * zc_pow[i] * SX;
            phi_tmp[256 + i] += 4.0 * xc_pow[32 + i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[64 + i] * zc_pow[32 + i] * SX;
            phi_tmp[288 + i] += 4.0 * xc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[320 + i] = xc_pow[32 + i] * yc_pow[64 + i] * SX;
            phi_tmp[320 + i] += 3.0 * xc_pow[i] * yc_pow[64 + i] * S0[i];

            phi_tmp[352 + i] = xc_pow[32 + i] * yc_pow[32 + i] * zc[i] * SX;
            phi_tmp[352 + i] += 3.0 * xc_pow[i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[384 + i] = xc_pow[32 + i] * yc_pow[i] * zc_pow[i] * SX;
            phi_tmp[384 + i] += 3.0 * xc_pow[i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[32 + i] * yc[i] * zc_pow[32 + i] * SX;
            phi_tmp[416 + i] += 3.0 * xc_pow[i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = xc_pow[32 + i] * zc_pow[64 + i] * SX;
            phi_tmp[448 + i] += 3.0 * xc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[480 + i] = xc_pow[i] * yc_pow[96 + i] * SX;
            phi_tmp[480 + i] += 2.0 * xc[i] * yc_pow[96 + i] * S0[i];

            phi_tmp[512 + i] = xc_pow[i] * yc_pow[64 + i] * zc[i] * SX;
            phi_tmp[512 + i] += 2.0 * xc[i] * yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[544 + i] = xc_pow[i] * yc_pow[32 + i] * zc_pow[i] * SX;
            phi_tmp[544 + i] += 2.0 * xc[i] * yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[576 + i] = xc_pow[i] * yc_pow[i] * zc_pow[32 + i] * SX;
            phi_tmp[576 + i] += 2.0 * xc[i] * yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[608 + i] = xc_pow[i] * yc[i] * zc_pow[64 + i] * SX;
            phi_tmp[608 + i] += 2.0 * xc[i] * yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[640 + i] = xc_pow[i] * zc_pow[96 + i] * SX;
            phi_tmp[640 + i] += 2.0 * xc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[672 + i] = xc[i] * yc_pow[128 + i] * SX;
            phi_tmp[672 + i] += yc_pow[128 + i] * S0[i];

            phi_tmp[704 + i] = xc[i] * yc_pow[96 + i] * zc[i] * SX;
            phi_tmp[704 + i] += yc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[736 + i] = xc[i] * yc_pow[64 + i] * zc_pow[i] * SX;
            phi_tmp[736 + i] += yc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[768 + i] = xc[i] * yc_pow[32 + i] * zc_pow[32 + i] * SX;
            phi_tmp[768 + i] += yc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[800 + i] = xc[i] * yc_pow[i] * zc_pow[64 + i] * SX;
            phi_tmp[800 + i] += yc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[832 + i] = xc[i] * yc[i] * zc_pow[96 + i] * SX;
            phi_tmp[832 + i] += yc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[864 + i] = xc[i] * zc_pow[128 + i] * SX;
            phi_tmp[864 + i] += zc_pow[128 + i] * S0[i];

            phi_tmp[896 + i] = yc_pow[160 + i] * SX;

            phi_tmp[928 + i] = yc_pow[128 + i] * zc[i] * SX;

            phi_tmp[960 + i] = yc_pow[96 + i] * zc_pow[i] * SX;

            phi_tmp[992 + i] = yc_pow[64 + i] * zc_pow[32 + i] * SX;

            phi_tmp[1024 + i] = yc_pow[32 + i] * zc_pow[64 + i] * SX;

            phi_tmp[1056 + i] = yc_pow[i] * zc_pow[96 + i] * SX;

            phi_tmp[1088 + i] = yc[i] * zc_pow[128 + i] * SX;

            phi_tmp[1120 + i] = zc_pow[160 + i] * SX;

        }

        if (spherical) {
            gg_cart_to_spherical_L7(remain, phi_tmp, 32, (phi_x_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_x_out + start), npoints, 0);
        }

        // Combine Y blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SY = S1[i] * yc[i];

            phi_tmp[i] = xc_pow[160 + i] * SY;

            phi_tmp[32 + i] = xc_pow[128 + i] * yc[i] * SY;
            phi_tmp[32 + i] += xc_pow[128 + i] * S0[i];

            phi_tmp[64 + i] = xc_pow[128 + i] * zc[i] * SY;

            phi_tmp[96 + i] = xc_pow[96 + i] * yc_pow[i] * SY;
            phi_tmp[96 + i] += 2.0 * xc_pow[96 + i] * yc[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[96 + i] * yc[i] * zc[i] * SY;
            phi_tmp[128 + i] += xc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[96 + i] * zc_pow[i] * SY;

            phi_tmp[192 + i] = xc_pow[64 + i] * yc_pow[32 + i] * SY;
            phi_tmp[192 + i] += 3.0 * xc_pow[64 + i] * yc_pow[i] * S0[i];

            phi_tmp[224 + i] = xc_pow[64 + i] * yc_pow[i] * zc[i] * SY;
            phi_tmp[224 + i] += 2.0 * xc_pow[64 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[64 + i] * yc[i] * zc_pow[i] * SY;
            phi_tmp[256 + i] += xc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[64 + i] * zc_pow[32 + i] * SY;

            phi_tmp[320 + i] = xc_pow[32 + i] * yc_pow[64 + i] * SY;
            phi_tmp[320 + i] += 4.0 * xc_pow[32 + i] * yc_pow[32 + i] * S0[i];

            phi_tmp[352 + i] = xc_pow[32 + i] * yc_pow[32 + i] * zc[i] * SY;
            phi_tmp[352 + i] += 3.0 * xc_pow[32 + i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[384 + i] = xc_pow[32 + i] * yc_pow[i] * zc_pow[i] * SY;
            phi_tmp[384 + i] += 2.0 * xc_pow[32 + i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[32 + i] * yc[i] * zc_pow[32 + i] * SY;
            phi_tmp[416 + i] += xc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = xc_pow[32 + i] * zc_pow[64 + i] * SY;

            phi_tmp[480 + i] = xc_pow[i] * yc_pow[96 + i] * SY;
            phi_tmp[480 + i] += 5.0 * xc_pow[i] * yc_pow[64 + i] * S0[i];

            phi_tmp[512 + i] = xc_pow[i] * yc_pow[64 + i] * zc[i] * SY;
            phi_tmp[512 + i] += 4.0 * xc_pow[i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[544 + i] = xc_pow[i] * yc_pow[32 + i] * zc_pow[i] * SY;
            phi_tmp[544 + i] += 3.0 * xc_pow[i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[576 + i] = xc_pow[i] * yc_pow[i] * zc_pow[32 + i] * SY;
            phi_tmp[576 + i] += 2.0 * xc_pow[i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[608 + i] = xc_pow[i] * yc[i] * zc_pow[64 + i] * SY;
            phi_tmp[608 + i] += xc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[640 + i] = xc_pow[i] * zc_pow[96 + i] * SY;

            phi_tmp[672 + i] = xc[i] * yc_pow[128 + i] * SY;
            phi_tmp[672 + i] += 6.0 * xc[i] * yc_pow[96 + i] * S0[i];

            phi_tmp[704 + i] = xc[i] * yc_pow[96 + i] * zc[i] * SY;
            phi_tmp[704 + i] += 5.0 * xc[i] * yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[736 + i] = xc[i] * yc_pow[64 + i] * zc_pow[i] * SY;
            phi_tmp[736 + i] += 4.0 * xc[i] * yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[768 + i] = xc[i] * yc_pow[32 + i] * zc_pow[32 + i] * SY;
            phi_tmp[768 + i] += 3.0 * xc[i] * yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[800 + i] = xc[i] * yc_pow[i] * zc_pow[64 + i] * SY;
            phi_tmp[800 + i] += 2.0 * xc[i] * yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[832 + i] = xc[i] * yc[i] * zc_pow[96 + i] * SY;
            phi_tmp[832 + i] += xc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[864 + i] = xc[i] * zc_pow[128 + i] * SY;

            phi_tmp[896 + i] = yc_pow[160 + i] * SY;
            phi_tmp[896 + i] += 7.0 * yc_pow[128 + i] * S0[i];

            phi_tmp[928 + i] = yc_pow[128 + i] * zc[i] * SY;
            phi_tmp[928 + i] += 6.0 * yc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[960 + i] = yc_pow[96 + i] * zc_pow[i] * SY;
            phi_tmp[960 + i] += 5.0 * yc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[992 + i] = yc_pow[64 + i] * zc_pow[32 + i] * SY;
            phi_tmp[992 + i] += 4.0 * yc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[1024 + i] = yc_pow[32 + i] * zc_pow[64 + i] * SY;
            phi_tmp[1024 + i] += 3.0 * yc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[1056 + i] = yc_pow[i] * zc_pow[96 + i] * SY;
            phi_tmp[1056 + i] += 2.0 * yc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[1088 + i] = yc[i] * zc_pow[128 + i] * SY;
            phi_tmp[1088 + i] += zc_pow[128 + i] * S0[i];

            phi_tmp[1120 + i] = zc_pow[160 + i] * SY;

        }

        if (spherical) {
            gg_cart_to_spherical_L7(remain, phi_tmp, 32, (phi_y_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_y_out + start), npoints, 0);
        }

        // Combine Z blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SZ = S1[i] * zc[i];

            phi_tmp[i] = xc_pow[160 + i] * SZ;

            phi_tmp[32 + i] = xc_pow[128 + i] * yc[i] * SZ;

            phi_tmp[64 + i] = xc_pow[128 + i] * zc[i] * SZ;
            phi_tmp[64 + i] += xc_pow[128 + i] * S0[i];

            phi_tmp[96 + i] = xc_pow[96 + i] * yc_pow[i] * SZ;

            phi_tmp[128 + i] = xc_pow[96 + i] * yc[i] * zc[i] * SZ;
            phi_tmp[128 + i] += xc_pow[96 + i] * yc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[96 + i] * zc_pow[i] * SZ;
            phi_tmp[160 + i] += 2.0 * xc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[192 + i] = xc_pow[64 + i] * yc_pow[32 + i] * SZ;

            phi_tmp[224 + i] = xc_pow[64 + i] * yc_pow[i] * zc[i] * SZ;
            phi_tmp[224 + i] += xc_pow[64 + i] * yc_pow[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[64 + i] * yc[i] * zc_pow[i] * SZ;
            phi_tmp[256 + i] += 2.0 * xc_pow[64 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[64 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[288 + i] += 3.0 * xc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[320 + i] = xc_pow[32 + i] * yc_pow[64 + i] * SZ;

            phi_tmp[352 + i] = xc_pow[32 + i] * yc_pow[32 + i] * zc[i] * SZ;
            phi_tmp[352 + i] += xc_pow[32 + i] * yc_pow[32 + i] * S0[i];

            phi_tmp[384 + i] = xc_pow[32 + i] * yc_pow[i] * zc_pow[i] * SZ;
            phi_tmp[384 + i] += 2.0 * xc_pow[32 + i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[32 + i] * yc[i] * zc_pow[32 + i] * SZ;
            phi_tmp[416 + i] += 3.0 * xc_pow[32 + i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[448 + i] = xc_pow[32 + i] * zc_pow[64 + i] * SZ;
            phi_tmp[448 + i] += 4.0 * xc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[480 + i] = xc_pow[i] * yc_pow[96 + i] * SZ;

            phi_tmp[512 + i] = xc_pow[i] * yc_pow[64 + i] * zc[i] * SZ;
            phi_tmp[512 + i] += xc_pow[i] * yc_pow[64 + i] * S0[i];

            phi_tmp[544 + i] = xc_pow[i] * yc_pow[32 + i] * zc_pow[i] * SZ;
            phi_tmp[544 + i] += 2.0 * xc_pow[i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[576 + i] = xc_pow[i] * yc_pow[i] * zc_pow[32 + i] * SZ;
            phi_tmp[576 + i] += 3.0 * xc_pow[i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[608 + i] = xc_pow[i] * yc[i] * zc_pow[64 + i] * SZ;
            phi_tmp[608 + i] += 4.0 * xc_pow[i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[640 + i] = xc_pow[i] * zc_pow[96 + i] * SZ;
            phi_tmp[640 + i] += 5.0 * xc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[672 + i] = xc[i] * yc_pow[128 + i] * SZ;

            phi_tmp[704 + i] = xc[i] * yc_pow[96 + i] * zc[i] * SZ;
            phi_tmp[704 + i] += xc[i] * yc_pow[96 + i] * S0[i];

            phi_tmp[736 + i] = xc[i] * yc_pow[64 + i] * zc_pow[i] * SZ;
            phi_tmp[736 + i] += 2.0 * xc[i] * yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[768 + i] = xc[i] * yc_pow[32 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[768 + i] += 3.0 * xc[i] * yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[800 + i] = xc[i] * yc_pow[i] * zc_pow[64 + i] * SZ;
            phi_tmp[800 + i] += 4.0 * xc[i] * yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[832 + i] = xc[i] * yc[i] * zc_pow[96 + i] * SZ;
            phi_tmp[832 + i] += 5.0 * xc[i] * yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[864 + i] = xc[i] * zc_pow[128 + i] * SZ;
            phi_tmp[864 + i] += 6.0 * xc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[896 + i] = yc_pow[160 + i] * SZ;

            phi_tmp[928 + i] = yc_pow[128 + i] * zc[i] * SZ;
            phi_tmp[928 + i] += yc_pow[128 + i] * S0[i];

            phi_tmp[960 + i] = yc_pow[96 + i] * zc_pow[i] * SZ;
            phi_tmp[960 + i] += 2.0 * yc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[992 + i] = yc_pow[64 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[992 + i] += 3.0 * yc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[1024 + i] = yc_pow[32 + i] * zc_pow[64 + i] * SZ;
            phi_tmp[1024 + i] += 4.0 * yc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[1056 + i] = yc_pow[i] * zc_pow[96 + i] * SZ;
            phi_tmp[1056 + i] += 5.0 * yc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[1088 + i] = yc[i] * zc_pow[128 + i] * SZ;
            phi_tmp[1088 + i] += 6.0 * yc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[1120 + i] = zc_pow[160 + i] * SZ;
            phi_tmp[1120 + i] += 7.0 * zc_pow[128 + i] * S0[i];

        }

        if (spherical) {
            gg_cart_to_spherical_L7(remain, phi_tmp, 32, (phi_z_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_z_out + start), npoints, 0);
        }

    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free Power temporaries
    ALIGNED_FREE(xc_pow);
    ALIGNED_FREE(yc_pow);
    ALIGNED_FREE(zc_pow);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);

}

void gg_collocation_L8_deriv1(const unsigned long npoints, const double* PRAGMA_RESTRICT x, const double* PRAGMA_RESTRICT y, const double* PRAGMA_RESTRICT z, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int spherical, double* PRAGMA_RESTRICT phi_out, double* PRAGMA_RESTRICT phi_x_out, double* PRAGMA_RESTRICT phi_y_out, double* PRAGMA_RESTRICT phi_z_out) {

    // Sizing
    unsigned long nblocks = npoints / 32;
    nblocks += (npoints % 32) ? 1 : 0;
    const unsigned long ncart = 45;
    const unsigned long nspherical = 17;
    const unsigned long nout = spherical ? nspherical : ncart;

    // Allocate S temporaries, single block to stay on cache
    double* PRAGMA_RESTRICT cache_data = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    double* PRAGMA_RESTRICT xc = cache_data + 0;
    ASSUME_ALIGNED(xc, 64);
    double* PRAGMA_RESTRICT yc = cache_data + 32;
    ASSUME_ALIGNED(yc, 64);
    double* PRAGMA_RESTRICT zc = cache_data + 64;
    ASSUME_ALIGNED(zc, 64);
    double* PRAGMA_RESTRICT R2 = cache_data + 96;
    ASSUME_ALIGNED(R2, 64);
    double* PRAGMA_RESTRICT S0 = cache_data + 128;
    ASSUME_ALIGNED(S0, 64);
    double* PRAGMA_RESTRICT tmp1 = cache_data + 160;
    ASSUME_ALIGNED(tmp1, 64);
    double* PRAGMA_RESTRICT S1 = cache_data + 192;
    ASSUME_ALIGNED(S1, 64);

    // Allocate exponential temporaries
    double* PRAGMA_RESTRICT expn1 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));
    double* PRAGMA_RESTRICT expn2 = (double*)ALIGNED_MALLOC(64, nprim * sizeof(double));

    // Allocate power temporaries
    double* PRAGMA_RESTRICT xc_pow = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    ASSUME_ALIGNED(xc_pow, 64);
    double* PRAGMA_RESTRICT yc_pow = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    ASSUME_ALIGNED(yc_pow, 64);
    double* PRAGMA_RESTRICT zc_pow = (double*)ALIGNED_MALLOC(64, 224 * sizeof(double));
    ASSUME_ALIGNED(zc_pow, 64);

    // Allocate output temporaries
    double* PRAGMA_RESTRICT phi_tmp = (double*)ALIGNED_MALLOC(64, 1440 * sizeof(double));
    ASSUME_ALIGNED(phi_tmp, 64);

    // Declare doubles
    const double center_x = center[0];
    const double center_y = center[1];
    const double center_z = center[2];
    double A;
    double AX, AY, AZ;

    // Build negative exponents
    for (unsigned long i = 0; i < nprim; i++) {
        expn1[i] = -1.0 * exponents[i];
        expn2[i] = -2.0 * exponents[i];
    }

    // Start outer block loop
    for (unsigned long block = 0; block < nblocks; block++) {


        // Copy data into inner temps
        const unsigned long start = block * 32;
        const unsigned long remain = ((start + 32) > npoints) ? (npoints - start) : 32;

        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            xc[i] = x[start + i] - center_x;
            yc[i] = y[start + i] - center_y;
            zc[i] = z[start + i] - center_z;

            // Distance
            R2[i] = xc[i] * xc[i];
            R2[i] += yc[i] * yc[i];
            R2[i] += zc[i] * zc[i];

            // Zero out S tmps
            S0[i] = 0.0;
            S1[i] = 0.0;
        }

        // Start exponential block loop
        for (unsigned long n = 0; n < nprim; n++) {
            const double coef = coeffs[n];
            const double alpha_n1 = expn1[n];
            const double alpha_n2 = expn2[n];

            PRAGMA_VECTORIZE
            for (unsigned long i = 0; i < remain; i++) {
                const double width = alpha_n1 * R2[i];
                const double T1 = coef * exp(width);
                S0[i] += T1;
                const double T2 = alpha_n2 * T1;
                S1[i] += T2;
            }

        }

        // Build powers
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            // Cartesian derivs
            xc_pow[i] = xc[i] * xc[i];
            yc_pow[i] = yc[i] * yc[i];
            zc_pow[i] = zc[i] * zc[i];
            xc_pow[32 + i] = xc_pow[i] * xc[i];
            yc_pow[32 + i] = yc_pow[i] * yc[i];
            zc_pow[32 + i] = zc_pow[i] * zc[i];
            xc_pow[64 + i] = xc_pow[32 + i] * xc[i];
            yc_pow[64 + i] = yc_pow[32 + i] * yc[i];
            zc_pow[64 + i] = zc_pow[32 + i] * zc[i];
            xc_pow[96 + i] = xc_pow[64 + i] * xc[i];
            yc_pow[96 + i] = yc_pow[64 + i] * yc[i];
            zc_pow[96 + i] = zc_pow[64 + i] * zc[i];
            xc_pow[128 + i] = xc_pow[96 + i] * xc[i];
            yc_pow[128 + i] = yc_pow[96 + i] * yc[i];
            zc_pow[128 + i] = zc_pow[96 + i] * zc[i];
            xc_pow[160 + i] = xc_pow[128 + i] * xc[i];
            yc_pow[160 + i] = yc_pow[128 + i] * yc[i];
            zc_pow[160 + i] = zc_pow[128 + i] * zc[i];
            xc_pow[192 + i] = xc_pow[160 + i] * xc[i];
            yc_pow[192 + i] = yc_pow[160 + i] * yc[i];
            zc_pow[192 + i] = zc_pow[160 + i] * zc[i];
        }
        // Combine A blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {

            phi_tmp[i] = xc_pow[192 + i] * S0[i];
            phi_tmp[32 + i] = xc_pow[160 + i] * yc[i] * S0[i];
            phi_tmp[64 + i] = xc_pow[160 + i] * zc[i] * S0[i];
            phi_tmp[96 + i] = xc_pow[128 + i] * yc_pow[i] * S0[i];
            phi_tmp[128 + i] = xc_pow[128 + i] * yc[i] * zc[i] * S0[i];
            phi_tmp[160 + i] = xc_pow[128 + i] * zc_pow[i] * S0[i];
            phi_tmp[192 + i] = xc_pow[96 + i] * yc_pow[32 + i] * S0[i];
            phi_tmp[224 + i] = xc_pow[96 + i] * yc_pow[i] * zc[i] * S0[i];
            phi_tmp[256 + i] = xc_pow[96 + i] * yc[i] * zc_pow[i] * S0[i];
            phi_tmp[288 + i] = xc_pow[96 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[320 + i] = xc_pow[64 + i] * yc_pow[64 + i] * S0[i];
            phi_tmp[352 + i] = xc_pow[64 + i] * yc_pow[32 + i] * zc[i] * S0[i];
            phi_tmp[384 + i] = xc_pow[64 + i] * yc_pow[i] * zc_pow[i] * S0[i];
            phi_tmp[416 + i] = xc_pow[64 + i] * yc[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[448 + i] = xc_pow[64 + i] * zc_pow[64 + i] * S0[i];
            phi_tmp[480 + i] = xc_pow[32 + i] * yc_pow[96 + i] * S0[i];
            phi_tmp[512 + i] = xc_pow[32 + i] * yc_pow[64 + i] * zc[i] * S0[i];
            phi_tmp[544 + i] = xc_pow[32 + i] * yc_pow[32 + i] * zc_pow[i] * S0[i];
            phi_tmp[576 + i] = xc_pow[32 + i] * yc_pow[i] * zc_pow[32 + i] * S0[i];
            phi_tmp[608 + i] = xc_pow[32 + i] * yc[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[640 + i] = xc_pow[32 + i] * zc_pow[96 + i] * S0[i];
            phi_tmp[672 + i] = xc_pow[i] * yc_pow[128 + i] * S0[i];
            phi_tmp[704 + i] = xc_pow[i] * yc_pow[96 + i] * zc[i] * S0[i];
            phi_tmp[736 + i] = xc_pow[i] * yc_pow[64 + i] * zc_pow[i] * S0[i];
            phi_tmp[768 + i] = xc_pow[i] * yc_pow[32 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[800 + i] = xc_pow[i] * yc_pow[i] * zc_pow[64 + i] * S0[i];
            phi_tmp[832 + i] = xc_pow[i] * yc[i] * zc_pow[96 + i] * S0[i];
            phi_tmp[864 + i] = xc_pow[i] * zc_pow[128 + i] * S0[i];
            phi_tmp[896 + i] = xc[i] * yc_pow[160 + i] * S0[i];
            phi_tmp[928 + i] = xc[i] * yc_pow[128 + i] * zc[i] * S0[i];
            phi_tmp[960 + i] = xc[i] * yc_pow[96 + i] * zc_pow[i] * S0[i];
            phi_tmp[992 + i] = xc[i] * yc_pow[64 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[1024 + i] = xc[i] * yc_pow[32 + i] * zc_pow[64 + i] * S0[i];
            phi_tmp[1056 + i] = xc[i] * yc_pow[i] * zc_pow[96 + i] * S0[i];
            phi_tmp[1088 + i] = xc[i] * yc[i] * zc_pow[128 + i] * S0[i];
            phi_tmp[1120 + i] = xc[i] * zc_pow[160 + i] * S0[i];
            phi_tmp[1152 + i] = yc_pow[192 + i] * S0[i];
            phi_tmp[1184 + i] = yc_pow[160 + i] * zc[i] * S0[i];
            phi_tmp[1216 + i] = yc_pow[128 + i] * zc_pow[i] * S0[i];
            phi_tmp[1248 + i] = yc_pow[96 + i] * zc_pow[32 + i] * S0[i];
            phi_tmp[1280 + i] = yc_pow[64 + i] * zc_pow[64 + i] * S0[i];
            phi_tmp[1312 + i] = yc_pow[32 + i] * zc_pow[96 + i] * S0[i];
            phi_tmp[1344 + i] = yc_pow[i] * zc_pow[128 + i] * S0[i];
            phi_tmp[1376 + i] = yc[i] * zc_pow[160 + i] * S0[i];
            phi_tmp[1408 + i] = zc_pow[192 + i] * S0[i];
        }

        if (spherical) {
            gg_cart_to_spherical_L8(remain, phi_tmp, 32, (phi_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_out + start), npoints, 0);
        }

        // Combine X blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SX = S1[i] * xc[i];

            phi_tmp[i] = xc_pow[192 + i] * SX;
            phi_tmp[i] += 8.0 * xc_pow[160 + i] * S0[i];

            phi_tmp[32 + i] = xc_pow[160 + i] * yc[i] * SX;
            phi_tmp[32 + i] += 7.0 * xc_pow[128 + i] * yc[i] * S0[i];

            phi_tmp[64 + i] = xc_pow[160 + i] * zc[i] * SX;
            phi_tmp[64 + i] += 7.0 * xc_pow[128 + i] * zc[i] * S0[i];

            phi_tmp[96 + i] = xc_pow[128 + i] * yc_pow[i] * SX;
            phi_tmp[96 + i] += 6.0 * xc_pow[96 + i] * yc_pow[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[128 + i] * yc[i] * zc[i] * SX;
            phi_tmp[128 + i] += 6.0 * xc_pow[96 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[128 + i] * zc_pow[i] * SX;
            phi_tmp[160 + i] += 6.0 * xc_pow[96 + i] * zc_pow[i] * S0[i];

            phi_tmp[192 + i] = xc_pow[96 + i] * yc_pow[32 + i] * SX;
            phi_tmp[192 + i] += 5.0 * xc_pow[64 + i] * yc_pow[32 + i] * S0[i];

            phi_tmp[224 + i] = xc_pow[96 + i] * yc_pow[i] * zc[i] * SX;
            phi_tmp[224 + i] += 5.0 * xc_pow[64 + i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[96 + i] * yc[i] * zc_pow[i] * SX;
            phi_tmp[256 + i] += 5.0 * xc_pow[64 + i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[96 + i] * zc_pow[32 + i] * SX;
            phi_tmp[288 + i] += 5.0 * xc_pow[64 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[320 + i] = xc_pow[64 + i] * yc_pow[64 + i] * SX;
            phi_tmp[320 + i] += 4.0 * xc_pow[32 + i] * yc_pow[64 + i] * S0[i];

            phi_tmp[352 + i] = xc_pow[64 + i] * yc_pow[32 + i] * zc[i] * SX;
            phi_tmp[352 + i] += 4.0 * xc_pow[32 + i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[384 + i] = xc_pow[64 + i] * yc_pow[i] * zc_pow[i] * SX;
            phi_tmp[384 + i] += 4.0 * xc_pow[32 + i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[64 + i] * yc[i] * zc_pow[32 + i] * SX;
            phi_tmp[416 + i] += 4.0 * xc_pow[32 + i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = xc_pow[64 + i] * zc_pow[64 + i] * SX;
            phi_tmp[448 + i] += 4.0 * xc_pow[32 + i] * zc_pow[64 + i] * S0[i];

            phi_tmp[480 + i] = xc_pow[32 + i] * yc_pow[96 + i] * SX;
            phi_tmp[480 + i] += 3.0 * xc_pow[i] * yc_pow[96 + i] * S0[i];

            phi_tmp[512 + i] = xc_pow[32 + i] * yc_pow[64 + i] * zc[i] * SX;
            phi_tmp[512 + i] += 3.0 * xc_pow[i] * yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[544 + i] = xc_pow[32 + i] * yc_pow[32 + i] * zc_pow[i] * SX;
            phi_tmp[544 + i] += 3.0 * xc_pow[i] * yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[576 + i] = xc_pow[32 + i] * yc_pow[i] * zc_pow[32 + i] * SX;
            phi_tmp[576 + i] += 3.0 * xc_pow[i] * yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[608 + i] = xc_pow[32 + i] * yc[i] * zc_pow[64 + i] * SX;
            phi_tmp[608 + i] += 3.0 * xc_pow[i] * yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[640 + i] = xc_pow[32 + i] * zc_pow[96 + i] * SX;
            phi_tmp[640 + i] += 3.0 * xc_pow[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[672 + i] = xc_pow[i] * yc_pow[128 + i] * SX;
            phi_tmp[672 + i] += 2.0 * xc[i] * yc_pow[128 + i] * S0[i];

            phi_tmp[704 + i] = xc_pow[i] * yc_pow[96 + i] * zc[i] * SX;
            phi_tmp[704 + i] += 2.0 * xc[i] * yc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[736 + i] = xc_pow[i] * yc_pow[64 + i] * zc_pow[i] * SX;
            phi_tmp[736 + i] += 2.0 * xc[i] * yc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[768 + i] = xc_pow[i] * yc_pow[32 + i] * zc_pow[32 + i] * SX;
            phi_tmp[768 + i] += 2.0 * xc[i] * yc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[800 + i] = xc_pow[i] * yc_pow[i] * zc_pow[64 + i] * SX;
            phi_tmp[800 + i] += 2.0 * xc[i] * yc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[832 + i] = xc_pow[i] * yc[i] * zc_pow[96 + i] * SX;
            phi_tmp[832 + i] += 2.0 * xc[i] * yc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[864 + i] = xc_pow[i] * zc_pow[128 + i] * SX;
            phi_tmp[864 + i] += 2.0 * xc[i] * zc_pow[128 + i] * S0[i];

            phi_tmp[896 + i] = xc[i] * yc_pow[160 + i] * SX;
            phi_tmp[896 + i] += yc_pow[160 + i] * S0[i];

            phi_tmp[928 + i] = xc[i] * yc_pow[128 + i] * zc[i] * SX;
            phi_tmp[928 + i] += yc_pow[128 + i] * zc[i] * S0[i];

            phi_tmp[960 + i] = xc[i] * yc_pow[96 + i] * zc_pow[i] * SX;
            phi_tmp[960 + i] += yc_pow[96 + i] * zc_pow[i] * S0[i];

            phi_tmp[992 + i] = xc[i] * yc_pow[64 + i] * zc_pow[32 + i] * SX;
            phi_tmp[992 + i] += yc_pow[64 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[1024 + i] = xc[i] * yc_pow[32 + i] * zc_pow[64 + i] * SX;
            phi_tmp[1024 + i] += yc_pow[32 + i] * zc_pow[64 + i] * S0[i];

            phi_tmp[1056 + i] = xc[i] * yc_pow[i] * zc_pow[96 + i] * SX;
            phi_tmp[1056 + i] += yc_pow[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[1088 + i] = xc[i] * yc[i] * zc_pow[128 + i] * SX;
            phi_tmp[1088 + i] += yc[i] * zc_pow[128 + i] * S0[i];

            phi_tmp[1120 + i] = xc[i] * zc_pow[160 + i] * SX;
            phi_tmp[1120 + i] += zc_pow[160 + i] * S0[i];

            phi_tmp[1152 + i] = yc_pow[192 + i] * SX;

            phi_tmp[1184 + i] = yc_pow[160 + i] * zc[i] * SX;

            phi_tmp[1216 + i] = yc_pow[128 + i] * zc_pow[i] * SX;

            phi_tmp[1248 + i] = yc_pow[96 + i] * zc_pow[32 + i] * SX;

            phi_tmp[1280 + i] = yc_pow[64 + i] * zc_pow[64 + i] * SX;

            phi_tmp[1312 + i] = yc_pow[32 + i] * zc_pow[96 + i] * SX;

            phi_tmp[1344 + i] = yc_pow[i] * zc_pow[128 + i] * SX;

            phi_tmp[1376 + i] = yc[i] * zc_pow[160 + i] * SX;

            phi_tmp[1408 + i] = zc_pow[192 + i] * SX;

        }

        if (spherical) {
            gg_cart_to_spherical_L8(remain, phi_tmp, 32, (phi_x_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_x_out + start), npoints, 0);
        }

        // Combine Y blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SY = S1[i] * yc[i];

            phi_tmp[i] = xc_pow[192 + i] * SY;

            phi_tmp[32 + i] = xc_pow[160 + i] * yc[i] * SY;
            phi_tmp[32 + i] += xc_pow[160 + i] * S0[i];

            phi_tmp[64 + i] = xc_pow[160 + i] * zc[i] * SY;

            phi_tmp[96 + i] = xc_pow[128 + i] * yc_pow[i] * SY;
            phi_tmp[96 + i] += 2.0 * xc_pow[128 + i] * yc[i] * S0[i];

            phi_tmp[128 + i] = xc_pow[128 + i] * yc[i] * zc[i] * SY;
            phi_tmp[128 + i] += xc_pow[128 + i] * zc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[128 + i] * zc_pow[i] * SY;

            phi_tmp[192 + i] = xc_pow[96 + i] * yc_pow[32 + i] * SY;
            phi_tmp[192 + i] += 3.0 * xc_pow[96 + i] * yc_pow[i] * S0[i];

            phi_tmp[224 + i] = xc_pow[96 + i] * yc_pow[i] * zc[i] * SY;
            phi_tmp[224 + i] += 2.0 * xc_pow[96 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[96 + i] * yc[i] * zc_pow[i] * SY;
            phi_tmp[256 + i] += xc_pow[96 + i] * zc_pow[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[96 + i] * zc_pow[32 + i] * SY;

            phi_tmp[320 + i] = xc_pow[64 + i] * yc_pow[64 + i] * SY;
            phi_tmp[320 + i] += 4.0 * xc_pow[64 + i] * yc_pow[32 + i] * S0[i];

            phi_tmp[352 + i] = xc_pow[64 + i] * yc_pow[32 + i] * zc[i] * SY;
            phi_tmp[352 + i] += 3.0 * xc_pow[64 + i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[384 + i] = xc_pow[64 + i] * yc_pow[i] * zc_pow[i] * SY;
            phi_tmp[384 + i] += 2.0 * xc_pow[64 + i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[64 + i] * yc[i] * zc_pow[32 + i] * SY;
            phi_tmp[416 + i] += xc_pow[64 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[448 + i] = xc_pow[64 + i] * zc_pow[64 + i] * SY;

            phi_tmp[480 + i] = xc_pow[32 + i] * yc_pow[96 + i] * SY;
            phi_tmp[480 + i] += 5.0 * xc_pow[32 + i] * yc_pow[64 + i] * S0[i];

            phi_tmp[512 + i] = xc_pow[32 + i] * yc_pow[64 + i] * zc[i] * SY;
            phi_tmp[512 + i] += 4.0 * xc_pow[32 + i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[544 + i] = xc_pow[32 + i] * yc_pow[32 + i] * zc_pow[i] * SY;
            phi_tmp[544 + i] += 3.0 * xc_pow[32 + i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[576 + i] = xc_pow[32 + i] * yc_pow[i] * zc_pow[32 + i] * SY;
            phi_tmp[576 + i] += 2.0 * xc_pow[32 + i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[608 + i] = xc_pow[32 + i] * yc[i] * zc_pow[64 + i] * SY;
            phi_tmp[608 + i] += xc_pow[32 + i] * zc_pow[64 + i] * S0[i];

            phi_tmp[640 + i] = xc_pow[32 + i] * zc_pow[96 + i] * SY;

            phi_tmp[672 + i] = xc_pow[i] * yc_pow[128 + i] * SY;
            phi_tmp[672 + i] += 6.0 * xc_pow[i] * yc_pow[96 + i] * S0[i];

            phi_tmp[704 + i] = xc_pow[i] * yc_pow[96 + i] * zc[i] * SY;
            phi_tmp[704 + i] += 5.0 * xc_pow[i] * yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[736 + i] = xc_pow[i] * yc_pow[64 + i] * zc_pow[i] * SY;
            phi_tmp[736 + i] += 4.0 * xc_pow[i] * yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[768 + i] = xc_pow[i] * yc_pow[32 + i] * zc_pow[32 + i] * SY;
            phi_tmp[768 + i] += 3.0 * xc_pow[i] * yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[800 + i] = xc_pow[i] * yc_pow[i] * zc_pow[64 + i] * SY;
            phi_tmp[800 + i] += 2.0 * xc_pow[i] * yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[832 + i] = xc_pow[i] * yc[i] * zc_pow[96 + i] * SY;
            phi_tmp[832 + i] += xc_pow[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[864 + i] = xc_pow[i] * zc_pow[128 + i] * SY;

            phi_tmp[896 + i] = xc[i] * yc_pow[160 + i] * SY;
            phi_tmp[896 + i] += 7.0 * xc[i] * yc_pow[128 + i] * S0[i];

            phi_tmp[928 + i] = xc[i] * yc_pow[128 + i] * zc[i] * SY;
            phi_tmp[928 + i] += 6.0 * xc[i] * yc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[960 + i] = xc[i] * yc_pow[96 + i] * zc_pow[i] * SY;
            phi_tmp[960 + i] += 5.0 * xc[i] * yc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[992 + i] = xc[i] * yc_pow[64 + i] * zc_pow[32 + i] * SY;
            phi_tmp[992 + i] += 4.0 * xc[i] * yc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[1024 + i] = xc[i] * yc_pow[32 + i] * zc_pow[64 + i] * SY;
            phi_tmp[1024 + i] += 3.0 * xc[i] * yc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[1056 + i] = xc[i] * yc_pow[i] * zc_pow[96 + i] * SY;
            phi_tmp[1056 + i] += 2.0 * xc[i] * yc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[1088 + i] = xc[i] * yc[i] * zc_pow[128 + i] * SY;
            phi_tmp[1088 + i] += xc[i] * zc_pow[128 + i] * S0[i];

            phi_tmp[1120 + i] = xc[i] * zc_pow[160 + i] * SY;

            phi_tmp[1152 + i] = yc_pow[192 + i] * SY;
            phi_tmp[1152 + i] += 8.0 * yc_pow[160 + i] * S0[i];

            phi_tmp[1184 + i] = yc_pow[160 + i] * zc[i] * SY;
            phi_tmp[1184 + i] += 7.0 * yc_pow[128 + i] * zc[i] * S0[i];

            phi_tmp[1216 + i] = yc_pow[128 + i] * zc_pow[i] * SY;
            phi_tmp[1216 + i] += 6.0 * yc_pow[96 + i] * zc_pow[i] * S0[i];

            phi_tmp[1248 + i] = yc_pow[96 + i] * zc_pow[32 + i] * SY;
            phi_tmp[1248 + i] += 5.0 * yc_pow[64 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[1280 + i] = yc_pow[64 + i] * zc_pow[64 + i] * SY;
            phi_tmp[1280 + i] += 4.0 * yc_pow[32 + i] * zc_pow[64 + i] * S0[i];

            phi_tmp[1312 + i] = yc_pow[32 + i] * zc_pow[96 + i] * SY;
            phi_tmp[1312 + i] += 3.0 * yc_pow[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[1344 + i] = yc_pow[i] * zc_pow[128 + i] * SY;
            phi_tmp[1344 + i] += 2.0 * yc[i] * zc_pow[128 + i] * S0[i];

            phi_tmp[1376 + i] = yc[i] * zc_pow[160 + i] * SY;
            phi_tmp[1376 + i] += zc_pow[160 + i] * S0[i];

            phi_tmp[1408 + i] = zc_pow[192 + i] * SY;

        }

        if (spherical) {
            gg_cart_to_spherical_L8(remain, phi_tmp, 32, (phi_y_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_y_out + start), npoints, 0);
        }

        // Combine Z blocks
        PRAGMA_VECTORIZE
        for (unsigned long i = 0; i < remain; i++) {
            const double SZ = S1[i] * zc[i];

            phi_tmp[i] = xc_pow[192 + i] * SZ;

            phi_tmp[32 + i] = xc_pow[160 + i] * yc[i] * SZ;

            phi_tmp[64 + i] = xc_pow[160 + i] * zc[i] * SZ;
            phi_tmp[64 + i] += xc_pow[160 + i] * S0[i];

            phi_tmp[96 + i] = xc_pow[128 + i] * yc_pow[i] * SZ;

            phi_tmp[128 + i] = xc_pow[128 + i] * yc[i] * zc[i] * SZ;
            phi_tmp[128 + i] += xc_pow[128 + i] * yc[i] * S0[i];

            phi_tmp[160 + i] = xc_pow[128 + i] * zc_pow[i] * SZ;
            phi_tmp[160 + i] += 2.0 * xc_pow[128 + i] * zc[i] * S0[i];

            phi_tmp[192 + i] = xc_pow[96 + i] * yc_pow[32 + i] * SZ;

            phi_tmp[224 + i] = xc_pow[96 + i] * yc_pow[i] * zc[i] * SZ;
            phi_tmp[224 + i] += xc_pow[96 + i] * yc_pow[i] * S0[i];

            phi_tmp[256 + i] = xc_pow[96 + i] * yc[i] * zc_pow[i] * SZ;
            phi_tmp[256 + i] += 2.0 * xc_pow[96 + i] * yc[i] * zc[i] * S0[i];

            phi_tmp[288 + i] = xc_pow[96 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[288 + i] += 3.0 * xc_pow[96 + i] * zc_pow[i] * S0[i];

            phi_tmp[320 + i] = xc_pow[64 + i] * yc_pow[64 + i] * SZ;

            phi_tmp[352 + i] = xc_pow[64 + i] * yc_pow[32 + i] * zc[i] * SZ;
            phi_tmp[352 + i] += xc_pow[64 + i] * yc_pow[32 + i] * S0[i];

            phi_tmp[384 + i] = xc_pow[64 + i] * yc_pow[i] * zc_pow[i] * SZ;
            phi_tmp[384 + i] += 2.0 * xc_pow[64 + i] * yc_pow[i] * zc[i] * S0[i];

            phi_tmp[416 + i] = xc_pow[64 + i] * yc[i] * zc_pow[32 + i] * SZ;
            phi_tmp[416 + i] += 3.0 * xc_pow[64 + i] * yc[i] * zc_pow[i] * S0[i];

            phi_tmp[448 + i] = xc_pow[64 + i] * zc_pow[64 + i] * SZ;
            phi_tmp[448 + i] += 4.0 * xc_pow[64 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[480 + i] = xc_pow[32 + i] * yc_pow[96 + i] * SZ;

            phi_tmp[512 + i] = xc_pow[32 + i] * yc_pow[64 + i] * zc[i] * SZ;
            phi_tmp[512 + i] += xc_pow[32 + i] * yc_pow[64 + i] * S0[i];

            phi_tmp[544 + i] = xc_pow[32 + i] * yc_pow[32 + i] * zc_pow[i] * SZ;
            phi_tmp[544 + i] += 2.0 * xc_pow[32 + i] * yc_pow[32 + i] * zc[i] * S0[i];

            phi_tmp[576 + i] = xc_pow[32 + i] * yc_pow[i] * zc_pow[32 + i] * SZ;
            phi_tmp[576 + i] += 3.0 * xc_pow[32 + i] * yc_pow[i] * zc_pow[i] * S0[i];

            phi_tmp[608 + i] = xc_pow[32 + i] * yc[i] * zc_pow[64 + i] * SZ;
            phi_tmp[608 + i] += 4.0 * xc_pow[32 + i] * yc[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[640 + i] = xc_pow[32 + i] * zc_pow[96 + i] * SZ;
            phi_tmp[640 + i] += 5.0 * xc_pow[32 + i] * zc_pow[64 + i] * S0[i];

            phi_tmp[672 + i] = xc_pow[i] * yc_pow[128 + i] * SZ;

            phi_tmp[704 + i] = xc_pow[i] * yc_pow[96 + i] * zc[i] * SZ;
            phi_tmp[704 + i] += xc_pow[i] * yc_pow[96 + i] * S0[i];

            phi_tmp[736 + i] = xc_pow[i] * yc_pow[64 + i] * zc_pow[i] * SZ;
            phi_tmp[736 + i] += 2.0 * xc_pow[i] * yc_pow[64 + i] * zc[i] * S0[i];

            phi_tmp[768 + i] = xc_pow[i] * yc_pow[32 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[768 + i] += 3.0 * xc_pow[i] * yc_pow[32 + i] * zc_pow[i] * S0[i];

            phi_tmp[800 + i] = xc_pow[i] * yc_pow[i] * zc_pow[64 + i] * SZ;
            phi_tmp[800 + i] += 4.0 * xc_pow[i] * yc_pow[i] * zc_pow[32 + i] * S0[i];

            phi_tmp[832 + i] = xc_pow[i] * yc[i] * zc_pow[96 + i] * SZ;
            phi_tmp[832 + i] += 5.0 * xc_pow[i] * yc[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[864 + i] = xc_pow[i] * zc_pow[128 + i] * SZ;
            phi_tmp[864 + i] += 6.0 * xc_pow[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[896 + i] = xc[i] * yc_pow[160 + i] * SZ;

            phi_tmp[928 + i] = xc[i] * yc_pow[128 + i] * zc[i] * SZ;
            phi_tmp[928 + i] += xc[i] * yc_pow[128 + i] * S0[i];

            phi_tmp[960 + i] = xc[i] * yc_pow[96 + i] * zc_pow[i] * SZ;
            phi_tmp[960 + i] += 2.0 * xc[i] * yc_pow[96 + i] * zc[i] * S0[i];

            phi_tmp[992 + i] = xc[i] * yc_pow[64 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[992 + i] += 3.0 * xc[i] * yc_pow[64 + i] * zc_pow[i] * S0[i];

            phi_tmp[1024 + i] = xc[i] * yc_pow[32 + i] * zc_pow[64 + i] * SZ;
            phi_tmp[1024 + i] += 4.0 * xc[i] * yc_pow[32 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[1056 + i] = xc[i] * yc_pow[i] * zc_pow[96 + i] * SZ;
            phi_tmp[1056 + i] += 5.0 * xc[i] * yc_pow[i] * zc_pow[64 + i] * S0[i];

            phi_tmp[1088 + i] = xc[i] * yc[i] * zc_pow[128 + i] * SZ;
            phi_tmp[1088 + i] += 6.0 * xc[i] * yc[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[1120 + i] = xc[i] * zc_pow[160 + i] * SZ;
            phi_tmp[1120 + i] += 7.0 * xc[i] * zc_pow[128 + i] * S0[i];

            phi_tmp[1152 + i] = yc_pow[192 + i] * SZ;

            phi_tmp[1184 + i] = yc_pow[160 + i] * zc[i] * SZ;
            phi_tmp[1184 + i] += yc_pow[160 + i] * S0[i];

            phi_tmp[1216 + i] = yc_pow[128 + i] * zc_pow[i] * SZ;
            phi_tmp[1216 + i] += 2.0 * yc_pow[128 + i] * zc[i] * S0[i];

            phi_tmp[1248 + i] = yc_pow[96 + i] * zc_pow[32 + i] * SZ;
            phi_tmp[1248 + i] += 3.0 * yc_pow[96 + i] * zc_pow[i] * S0[i];

            phi_tmp[1280 + i] = yc_pow[64 + i] * zc_pow[64 + i] * SZ;
            phi_tmp[1280 + i] += 4.0 * yc_pow[64 + i] * zc_pow[32 + i] * S0[i];

            phi_tmp[1312 + i] = yc_pow[32 + i] * zc_pow[96 + i] * SZ;
            phi_tmp[1312 + i] += 5.0 * yc_pow[32 + i] * zc_pow[64 + i] * S0[i];

            phi_tmp[1344 + i] = yc_pow[i] * zc_pow[128 + i] * SZ;
            phi_tmp[1344 + i] += 6.0 * yc_pow[i] * zc_pow[96 + i] * S0[i];

            phi_tmp[1376 + i] = yc[i] * zc_pow[160 + i] * SZ;
            phi_tmp[1376 + i] += 7.0 * yc[i] * zc_pow[128 + i] * S0[i];

            phi_tmp[1408 + i] = zc_pow[192 + i] * SZ;
            phi_tmp[1408 + i] += 8.0 * zc_pow[160 + i] * S0[i];

        }

        if (spherical) {
            gg_cart_to_spherical_L8(remain, phi_tmp, 32, (phi_z_out + start), npoints);
            } else {
            block_copy(nout, remain, phi_tmp, 32, (phi_z_out + start), npoints, 0);
        }

    }

    // Free S temporaries
    ALIGNED_FREE(cache_data);
    ALIGNED_FREE(expn1);
    ALIGNED_FREE(expn2);

    // Free Power temporaries
    ALIGNED_FREE(xc_pow);
    ALIGNED_FREE(yc_pow);
    ALIGNED_FREE(zc_pow);

    // Free inner temporaries
    ALIGNED_FREE(phi_tmp);

}
